#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

import sys
from os import getcwd
from typing import Optional

import torch
from torch import optim

from tardis_em.dist_pytorch.dist import CDIST, DIST
from tardis_em.dist_pytorch.sparse_dist import SparseDIST
from tardis_em.dist_pytorch.trainer import (
    CDistTrainer,
    DistTrainer,
    SparseDistTrainer,
)
from tardis_em.dist_pytorch.utils.utils import check_model_dict
from tardis_em.utils.device import get_device
from tardis_em.utils.logo import TardisLogo
from tardis_em.utils.losses import (
    AdaptiveDiceLoss,
    BCELoss,
    BCEGraphWiseLoss,
    WBCELoss,
    BCEDiceLoss,
    CELoss,
    DiceLoss,
    ClDiceLoss,
    ClBCELoss,
    SigmoidFocalLoss,
    LaplacianEigenmapsLoss,
)
from tardis_em.utils.trainer import ISR_LR

# Setting for stable release to turn off all debug APIs
torch.backends.cudnn.benchmark = True
torch.autograd.set_detect_anomaly(mode=False)
torch.autograd.profiler.profile(enabled=False)
torch.autograd.profiler.emit_nvtx(enabled=False)


def train_dist(
    dataset_type: str,
    edge_angles: bool,
    train_dataloader,
    test_dataloader,
    model_structure: dict,
    checkpoint: Optional[str] = None,
    loss_function="bce",
    learning_rate=0.001,
    lr_scheduler=False,
    early_stop_rate=10,
    device="gpu",
    epochs=1000,
):
    """
    Wrapper for DIST or C_DIST models.

    Args:
        dataset_type (str): Type of input dataset.
        edge_angles(bool): If True, use an angle for embedding.
        train_dataloader (torch.DataLoader): DataLoader with train dataset.
        test_dataloader (torch.DataLoader): DataLoader with test dataset.
        model_structure (dict): Dictionary with model setting.
        checkpoint (None, optional): Optional, DIST model checkpoint.
        loss_function (str): Type of loss function.
        learning_rate (float): Learning rate.
        lr_scheduler (bool): If True, LR_scheduler is used with training.
        early_stop_rate (int): Define max. number of epoch's without improvements
        after which training is stopped.
        device (torch.device): Device on which model is trained.
        epochs (int): Max number of epoch's.
    """
    """Losses"""
    loss_functions = [
        AdaptiveDiceLoss,
        BCELoss,
        BCEGraphWiseLoss,
        WBCELoss,
        BCEDiceLoss,
        CELoss,
        DiceLoss,
        ClDiceLoss,
        ClBCELoss,
        SigmoidFocalLoss,
        LaplacianEigenmapsLoss,
    ]
    losses_f = {
        f.__name__: f(smooth=1e-16, reduction="mean", diagonal=False, sigmoid=False)
        for f in loss_functions
    }

    """Check input variable"""
    model_structure = check_model_dict(model_structure)

    if not isinstance(device, torch.device) and isinstance(device, str):
        device = get_device(device)

    """Build DIST model"""
    if model_structure["dist_type"] == "instance":
        model = DIST(
            n_out=model_structure["n_out"],
            node_input=model_structure["node_input"],
            node_dim=model_structure["node_dim"],
            edge_dim=model_structure["edge_dim"],
            num_layers=model_structure["num_layers"],
            num_heads=model_structure["num_heads"],
            rgb_embed_sigma=model_structure["rgb_embed_sigma"],
            coord_embed_sigma=model_structure["coord_embed_sigma"],
            dropout_rate=model_structure["dropout_rate"],
            structure=model_structure["structure"],
            predict=True,
            edge_angles=edge_angles,
        )
    elif model_structure["dist_type"] == "instance-sparse":
        losses_f = {
            f.__name__: f(smooth=1e-16, reduction="mean", diagonal=False, sigmoid=False)
            for f in loss_functions
        }
        model = SparseDIST(
            n_out=model_structure["n_out"],
            edge_dim=model_structure["edge_dim"],
            num_layers=model_structure["num_layers"],
            knn=model_structure["num_knn"],
            coord_embed_sigma=model_structure["coord_embed_sigma"],
            predict=True,
            device=device,
        )
    elif model_structure["dist_type"] == "semantic":
        model = CDIST(
            n_out=model_structure["n_out"],
            node_input=model_structure["node_input"],
            node_dim=model_structure["node_dim"],
            edge_dim=model_structure["edge_dim"],
            num_layers=model_structure["num_layers"],
            num_heads=model_structure["num_heads"],
            num_cls=model_structure["num_cls"],
            rgb_embed_sigma=model_structure["rgb_embed_sigma"],
            coord_embed_sigma=model_structure["coord_embed_sigma"],
            dropout_rate=model_structure["dropout_rate"],
            structure=model_structure["structure"],
            predict=True,
        )
    else:
        tardis_logo = TardisLogo()
        tardis_logo(
            text_1=f"ValueError: Model type: DIST-{model_structure['dist_type']} is not supported!"
        )
        sys.exit()

    """Build TARDIS progress bar output"""
    if model_structure["node_dim"] == 0:
        node_sigma = " "
    elif model_structure["rgb_embed_sigma"] == 0:
        node_sigma = ", [Linear] node_sigma"
    else:
        node_sigma = f", {model_structure['rgb_embed_sigma']} node_sigma"

    print_setting = [
        f"Training is started on {device} for DIST-{model_structure['structure']}",
        f"Local dir: {getcwd()}",
        f"Training for {model_structure['dist_type']} with "
        f"No. of Layers: {model_structure['num_layers']} and "
        f"{model_structure['num_heads']} heads",
        f"Layers are build of {model_structure['node_dim']} nodes, "
        f"{model_structure['edge_dim']} edges, "
        f"{model_structure['coord_embed_sigma']} edge_sigma{node_sigma}",
    ]

    """Optionally: Load checkpoint for retraining"""
    if checkpoint is not None:
        save_train = torch.load(checkpoint, map_location=device)

        if "model_struct_dict" in save_train.keys():
            model_dict = save_train["model_struct_dict"]
            globals().update(model_dict)

        model.load_state_dict(save_train["model_state_dict"])

    model = model.to(device)

    """Define loss function for training"""
    loss_fn = losses_f["BCELoss"]
    if loss_function in losses_f:
        loss_fn = losses_f[loss_function]

    """Build training optimizer"""
    if lr_scheduler:
        optimizer = optim.Adam(params=model.parameters(), betas=(0.9, 0.98), eps=1e-9)
        optimizer = ISR_LR(optimizer, lr_mul=learning_rate)
    else:
        optimizer = optim.Adam(
            params=model.parameters(), lr=learning_rate, betas=(0.9, 0.98), eps=1e-9
        )

    """Optionally: Checkpoint model"""
    if checkpoint is not None:
        optimizer.load_state_dict(save_train["optimizer_state_dict"])
        del save_train

    """Build trainer"""
    if dataset_type in ["filament", "MT", "Mem"] or dataset_type[1] in [
        "filament",
        "membrane2d",
        "mix2d",
    ]:
        dataset_type = 2
    else:
        dataset_type = 4

    if model_structure["dist_type"] == "instance":
        train = DistTrainer(
            model=model,
            structure=model_structure,
            instance_cov=dataset_type,
            device=device,
            criterion=loss_fn,
            optimizer=optimizer,
            print_setting=print_setting,
            training_DataLoader=train_dataloader,
            validation_DataLoader=test_dataloader,
            epochs=epochs,
            lr_scheduler=lr_scheduler,
            early_stop_rate=early_stop_rate,
            checkpoint_name=model_structure["dist_type"],
        )
    elif model_structure["dist_type"] == "instance-sparse":
        train = SparseDistTrainer(
            model=model,
            structure=model_structure,
            instance_cov=dataset_type,
            device=device,
            criterion=loss_fn,
            optimizer=optimizer,
            print_setting=print_setting,
            training_DataLoader=train_dataloader,
            validation_DataLoader=test_dataloader,
            epochs=epochs,
            lr_scheduler=lr_scheduler,
            early_stop_rate=early_stop_rate,
            checkpoint_name=model_structure["dist_type"],
        )
    else:
        train = CDistTrainer(
            model=model,
            structure=model_structure,
            device=device,
            criterion=loss_fn,
            optimizer=optimizer,
            print_setting=print_setting,
            training_DataLoader=train_dataloader,
            validation_DataLoader=test_dataloader,
            epochs=epochs,
            lr_scheduler=lr_scheduler,
            early_stop_rate=early_stop_rate,
            checkpoint_name=model_structure["dist_type"],
        )

    """Train"""
    train.run_trainer()



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
from math import sqrt
from typing import Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F


class PairBiasSelfAttention(nn.Module):
    """
    SELF-ATTENTION WITH EDGE FEATURE-BASED BIAS AND PRE-LAYER NORMALIZATION

    Self-attention block that attends coordinate and image patches or RGB.
    """

    def __init__(
        self,
        embed_dim: int,
        pairs_dim: int,
        num_heads: int,
        init_scaling=1 / 1.4142135623730951,
    ):
        """

        Args:
            embed_dim (int): Number of embedded dimensions for node dimensions.
            pairs_dim (int): Number of pairs dimensions.
            num_heads (int): Number of heads for multi-head attention.
            init_scaling (float): Initial scaling factor used for reset parameters.
        """
        super().__init__()
        # Embedding setting
        self.embed_dim = self.kdim = self.vdim = embed_dim
        self.qkv_same_dim = self.kdim == embed_dim and self.vdim == embed_dim

        self.num_heads = num_heads
        self.head_dim = embed_dim // num_heads
        assert (
            self.head_dim * num_heads == self.embed_dim
        ), "embed_dim must be divisible by num_heads"
        self.scaling = self.head_dim**-0.5

        # Initializing scaling factor
        self.init_scaling = init_scaling

        # Embedding linear layers
        self.norm_input = nn.LayerNorm(embed_dim)
        self.q_proj = nn.Linear(embed_dim, embed_dim, bias=False)
        self.k_proj = nn.Linear(self.kdim, embed_dim, bias=False)
        self.v_proj = nn.Linear(self.vdim, embed_dim, bias=False)
        self.out_proj = nn.Linear(embed_dim, embed_dim)

        self.pairs_dim = pairs_dim
        self.norm_pairs = nn.LayerNorm(pairs_dim)
        self.pairs_proj = nn.Linear(self.pairs_dim, self.num_heads, bias=False)

        self._reset_parameters()

    def _reset_parameters(self):
        """
        Initial parameter and bias scaling.
        """
        nn.init.xavier_uniform_(self.q_proj.weight, gain=self.init_scaling)
        nn.init.xavier_uniform_(self.k_proj.weight, gain=self.init_scaling)
        nn.init.xavier_uniform_(self.v_proj.weight, gain=self.init_scaling)

        nn.init.constant_(self.out_proj.weight, 0.0)
        if self.out_proj.bias is not None:
            nn.init.constant_(self.out_proj.bias, 0.0)

        nn.init.xavier_uniform_(self.pairs_proj.weight, gain=self.init_scaling)

    def forward(
        self,
        query: torch.Tensor,
        pairs: torch.Tensor,
        attn_mask: Optional[torch.Tensor] = None,
        key_padding_mask: Optional[torch.Tensor] = None,
        need_weights: bool = False,
        need_head_weights: bool = False,
    ) -> Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:
        """
        Forward attention over node features.

        Args:
            query (torch.Tensor): Nodes features  [Length x Batch x Channel].
            pairs (torch.Tensor): Edges features [Batch x Length x Length x Channel].
            attn_mask (torch.Tensor): Typically used to implement causal attention,
                where the mask prevents the attention from looking forward in time.
            key_padding_mask (torch.Tensor): Mask to exclude keys that are pads,
                of shape [Batch, src_len]
            need_weights (bool): If True, return the attention weights,
                and averaged overheads.
            need_head_weights (bool): If True, return the attention weights
                for each head.

        Returns:
            Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor]: Attention tensor
            for node features.
        """
        if need_head_weights:
            need_weights = True

        tgt_len, bsz, embed_dim = query.size()
        assert embed_dim == self.embed_dim
        assert list(query.size()) == [tgt_len, bsz, embed_dim]

        query = self.norm_input(query)
        q = self.q_proj(query)
        k = self.k_proj(query)
        v = self.v_proj(query)

        q *= self.scaling
        q = (
            q.contiguous()
            .view(tgt_len, bsz * self.num_heads, self.head_dim)
            .transpose(0, 1)
        )
        k = k.contiguous().view(-1, bsz * self.num_heads, self.head_dim).transpose(0, 1)
        v = v.contiguous().view(-1, bsz * self.num_heads, self.head_dim).transpose(0, 1)
        src_len = k.size(1)
        """
        This is part of a workaround to get around fork/join parallelism
        not supporting Optional types.
        """
        if key_padding_mask is not None and key_padding_mask.dim() == 0:
            key_padding_mask = None

        if key_padding_mask is not None:
            assert key_padding_mask.size(0) == bsz
            assert key_padding_mask.size(1) == src_len

        attn_weights = torch.bmm(q, k.transpose(1, 2))
        assert list(attn_weights.size()) == [bsz * self.num_heads, tgt_len, src_len]

        pairs = self.norm_pairs(pairs)
        pair_weights = self.pairs_proj(pairs)

        # pair_weights are B x L x L x H, needs to be B x H x L x L
        pair_weights = pair_weights.permute(0, 3, 1, 2)
        attn_weights = (
            attn_weights.view(bsz, self.num_heads, tgt_len, src_len) + pair_weights
        )
        attn_weights = attn_weights.view(bsz * self.num_heads, tgt_len, src_len)

        if attn_mask is not None:
            attn_mask = attn_mask.unsqueeze(0)
            attn_weights += attn_mask

        if key_padding_mask is not None:
            # don't attend to padding symbols
            attn_weights = attn_weights.view(bsz, self.num_heads, tgt_len, src_len)

            key_padding_mask = key_padding_mask.unsqueeze(1).unsqueeze(2).to(torch.bool)
            attn_weights = attn_weights.masked_fill(key_padding_mask, float("-inf"))

            attn_weights = attn_weights.view(bsz * self.num_heads, tgt_len, src_len)

        attn_weights_float = F.softmax(attn_weights, dim=-1)
        attn_weights = attn_weights_float.type_as(attn_weights)
        attn_probs = attn_weights

        assert v is not None
        attn = torch.bmm(attn_probs, v)
        assert list(attn.size()) == [bsz * self.num_heads, tgt_len, self.head_dim]
        attn = attn.transpose(0, 1).contiguous().view(tgt_len, bsz, embed_dim)
        attn = self.out_proj(attn)

        if need_weights:
            attn_weights = attn_weights_float.view(
                bsz, self.num_heads, tgt_len, src_len
            )
            if not need_head_weights:
                # average attention weights over heads
                attn_weights = attn_weights.mean(dim=1)

            return attn, attn_weights
        return attn


class ComparisonLayer(nn.Module):
    """
    COMPARISON MODULE BETWEEN PAIRS AND NODE INPUTS

    This module converts pairs representation of dim (Length x Batch x Channels)
    into (Batch x Length x Length x Channels) representation that can be compared
    with node representation.

    Args:
        input_dim (int): Input dimension as in pairs features.
        output_dim (int): Output dimension as in node features.
        channel_dim (int): Number of output channels.
    """

    def __init__(self, input_dim: int, output_dim: int, channel_dim=128):
        super().__init__()
        # Linear layer for converting pairs node futures to edge shape.
        self.norm = nn.LayerNorm(input_dim)
        self.linear1 = nn.Linear(input_dim, channel_dim)
        self.linear2 = nn.Linear(input_dim, channel_dim)
        self.linear3 = nn.Linear(channel_dim, output_dim)
        self.linear4 = nn.Linear(channel_dim, output_dim, bias=False)

        # Parameter initialization
        nn.init.constant_(self.linear3.weight, 0.0)
        nn.init.constant_(self.linear3.bias, 0.0)

        nn.init.constant_(self.linear4.weight, 0.0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward node compatible layer.

        Args:
            x (torch.Tensor): Node features after attention layer.

        Returns:
            torch.Tensor: Converted Node features to [Batch x Length x Length x
            Out_Channels] shape.
        """
        x = x.transpose(0, 1)
        x = self.norm(x)
        a = self.linear1(x)
        b = self.linear2(x)

        """Batch x Length x Length x Out_Channels"""
        return self.linear3(a.unsqueeze(2) * b.unsqueeze(1)) + self.linear4(
            a.unsqueeze(2) - b.unsqueeze(1)
        )


class TriangularEdgeUpdate(nn.Module):
    """
    TRIANGULAR UPDATE MODEL FOR NODES FEATURES

    This module takes node feature representation and performs triangular
    attention for each point. Similar to in Alphafold2 approach.

    Args:
        input_dim (int): Number of input channels.
        channel_dim (int): Number of output channels.
        axis (int): Indicate the axis around which the attention is given.
    """

    def __init__(self, input_dim, channel_dim=128, axis=1):
        super().__init__()
        self.input_dim = input_dim
        self.channel_dim = channel_dim
        self.axis = axis
        self.init_scaling = 1 / sqrt(2)

        self.norm_input = nn.LayerNorm(input_dim)

        self.linear_a = nn.Linear(input_dim, channel_dim)
        self.gate_a = nn.Linear(input_dim, channel_dim)

        self.linear_b = nn.Linear(input_dim, channel_dim)
        self.gate_b = nn.Linear(input_dim, channel_dim)

        self.norm_o = nn.LayerNorm(channel_dim)
        self.gate_o = nn.Linear(input_dim, input_dim)
        self.linear_o = nn.Linear(channel_dim, input_dim)

        self._reset_parameters()

    def _reset_parameters(self):
        """
        Initial parameter and bias scaling.
        """
        nn.init.xavier_uniform_(self.linear_a.weight, gain=self.init_scaling)
        nn.init.constant_(self.linear_a.bias, 0.0)

        nn.init.xavier_uniform_(self.linear_b.weight, gain=self.init_scaling)
        nn.init.constant_(self.linear_b.bias, 0.0)

        nn.init.constant_(self.linear_o.weight, 0.0)
        nn.init.constant_(self.linear_o.bias, 0.0)

    def forward(
        self, z: torch.Tensor, mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Forward Triangular edge update.

        Args:
            z (torch.Tensor): Edge features.
            mask (torch.Tensor, optional): Optional mask torch.Tensor layer.

        Returns:
            torch.Tensor: Updated edge features.
        """
        z = self.norm_input(z)

        a = torch.sigmoid(self.gate_a(z)) * self.linear_a(z)  # B x L x L x O
        b = torch.sigmoid(self.gate_b(z)) * self.linear_b(z)  # B x L x L x O

        if mask is not None:
            mask = mask.unsqueeze(3).expand(
                mask.size(0), mask.size(1), mask.size(2), a.size(3)
            )
            a = torch.where(mask, torch.zeros_like(a), a)
            b = torch.where(mask, torch.zeros_like(b), b)

        # i,j -> i,k j,k
        if self.axis == 1:
            k = torch.einsum("biko,bjko->bijo", a, b)
        else:
            k = torch.einsum("bkio,bkjo->bijo", a, b)

        return torch.sigmoid(self.gate_o(z)) * self.linear_o(self.norm_o(k))


class QuadraticEdgeUpdate(nn.Module):
    """
    QUADRATIC UPDATE MODEL FOR NODES FEATURES

    This module takes node feature representation and performs quadratic
    attention for each point. This is a modified Alphafold2 solution.

    Args:
        input_dim (int): Number of input channels.
        channel_dim (int): Number of output channels.
        axis (int): Indicate the axis around which the attention is given.
    """

    def __init__(self, input_dim, channel_dim=128, axis=1):
        super().__init__()
        self.input_dim = input_dim
        self.channel_dim = channel_dim
        self.axis = axis
        self.init_scaling = 1 / sqrt(2)

        self.norm_input = nn.LayerNorm(input_dim)

        self.linear_a = nn.Linear(input_dim, channel_dim)
        self.gate_a = nn.Linear(input_dim, channel_dim)

        self.linear_b = nn.Linear(input_dim, channel_dim)
        self.gate_b = nn.Linear(input_dim, channel_dim)

        self.linear_c = nn.Linear(input_dim, channel_dim)
        self.gate_c = nn.Linear(input_dim, channel_dim)

        self.linear_d = nn.Linear(input_dim, channel_dim)
        self.gate_d = nn.Linear(input_dim, channel_dim)

        self.norm_o = nn.LayerNorm(channel_dim)
        self.gate_o = nn.Linear(input_dim, input_dim)
        self.linear_o = nn.Linear(channel_dim, input_dim)

        self._reset_parameters()

    def _reset_parameters(self):
        """
        Initial parameter and bias scaling.
        """
        nn.init.xavier_uniform_(self.linear_a.weight, gain=self.init_scaling)
        nn.init.constant_(self.linear_a.bias, 0.0)

        nn.init.xavier_uniform_(self.linear_b.weight, gain=self.init_scaling)
        nn.init.constant_(self.linear_b.bias, 0.0)

        nn.init.xavier_uniform_(self.linear_c.weight, gain=self.init_scaling)
        nn.init.constant_(self.linear_c.bias, 0.0)

        nn.init.xavier_uniform_(self.linear_d.weight, gain=self.init_scaling)
        nn.init.constant_(self.linear_d.bias, 0.0)

        nn.init.constant_(self.linear_o.weight, 0.0)
        nn.init.constant_(self.linear_o.bias, 0.0)

    def forward(
        self, z: torch.Tensor, mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Forward Quadratic edge update.

        Args:
            z (torch.Tensor): Edge features.
            mask (torch.Tensor, optional): Optional mask torch.Tensor layer.

        Returns:
            torch.Tensor: Updated edge features.
        """
        z = self.norm_input(z)

        a = torch.sigmoid(self.gate_a(z)) * self.linear_a(z)  # B x L x L x O
        b = torch.sigmoid(self.gate_b(z)) * self.linear_b(z)  # B x L x L x O
        c = torch.sigmoid(self.gate_c(z)) * self.linear_c(z)  # B x L x L x O
        d = torch.sigmoid(self.gate_d(z)) * self.linear_d(z)  # B x L x L x O

        if mask is not None:
            mask = mask.unsqueeze(3).expand(
                mask.size(0), mask.size(1), mask.size(2), a.size(3)
            )
            a = torch.where(mask, torch.zeros_like(a), a)
            b = torch.where(mask, torch.zeros_like(b), b)
            c = torch.where(mask, torch.zeros_like(c), c)
            d = torch.where(mask, torch.zeros_like(d), d)

        # i,j -> i,k j,k, il, jl
        if self.axis == 1:
            k = torch.einsum("biko,bjko,bilo,bjlo->bijo", a, b, c, d)
        else:
            k = torch.einsum("bkio,bkjo,blio,bljo->bijo", a, b, c, d)

        return torch.sigmoid(self.gate_o(z)) * self.linear_o(self.norm_o(k))


class MultiHeadAttention(nn.Module):
    """
    MULTI-HEADED ATTENTION

    See "Attention Is All You Need" for more details. Modified from 'fairseq'.

    Args:
        embed_dim (int): Number of embedded dimensions for node features.
        num_heads (int): Number of heads for multi-head attention.
        kdim: Key dimensions.
        vdim: Values dimensions.
        dropout (float): Dropout probability.
        bias (bool): If True add bias.
        add_bias_kv (bool): If True add bias for keys and values.
        add_zero_attn (bool): If True replace attention with a zero-out mask.
        self_attention (bool): If True self-attention is used.
        encoder_decoder_attention (bool): If True self-attention over
            encode/decoder is used.
        init_scaling (float): The initial scaling factor used for reset parameters.
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        kdim=None,
        vdim=None,
        dropout=0.0,
        bias=True,
        add_bias_kv=False,
        add_zero_attn=False,
        self_attention=False,
        encoder_decoder_attention=False,
        init_scaling=1 / sqrt(2),
    ):
        super().__init__()

        self.embed_dim = embed_dim
        self.kdim = kdim if kdim is not None else embed_dim
        self.vdim = vdim if vdim is not None else embed_dim
        self.qkv_same_dim = self.kdim == embed_dim and self.vdim == embed_dim

        self.num_heads = num_heads
        self.dropout_module = nn.Dropout(dropout)

        self.head_dim = embed_dim // num_heads
        assert (
            self.head_dim * num_heads == self.embed_dim
        ), "embed_dim must be divisible by num_heads"
        self.scaling = self.head_dim**-0.5

        self.self_attention = self_attention
        self.encoder_decoder_attention = encoder_decoder_attention

        assert (
            not self.self_attention or self.qkv_same_dim
        ), "Self-attention requires query, key and value to be of the same size"

        self.init_scaling = init_scaling
        self.v_proj = nn.Linear(self.vdim, embed_dim, bias=bias)
        self.q_proj = nn.Linear(embed_dim, embed_dim, bias=bias)
        self.k_proj = nn.Linear(self.kdim, embed_dim, bias=bias)
        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias)

        if add_bias_kv:
            self.bias_k = nn.Parameter(torch.Tensor(1, 1, embed_dim))
            self.bias_v = nn.Parameter(torch.Tensor(1, 1, embed_dim))
        else:
            self.bias_k = self.bias_v = None

        self.add_zero_attn = add_zero_attn

        self.reset_parameters()

        self.onnx_trace = False
        self.tpu = False

    def reset_parameters(self):
        """
        Initial parameter and bias scaling.
        """
        nn.init.xavier_uniform_(self.k_proj.weight, gain=self.init_scaling)
        nn.init.xavier_uniform_(self.v_proj.weight, gain=self.init_scaling)
        nn.init.xavier_uniform_(self.q_proj.weight, gain=self.init_scaling)
        nn.init.constant_(self.out_proj.weight, 0.0)

        if self.out_proj.bias is not None:
            nn.init.constant_(self.out_proj.bias, 0.0)
        if self.bias_k is not None:
            nn.init.constant_(self.bias_k.bias, 0.0)
        if self.bias_v is not None:
            nn.init.constant_(self.bias_v.bias, 0.0)

    def forward(
        self,
        query: torch.Tensor,
        key: Optional[torch.Tensor] = None,
        value: Optional[torch.Tensor] = None,
        key_padding_mask: Optional[torch.Tensor] = None,
        need_weights: bool = False,
        attn_mask: Optional[torch.Tensor] = None,
        before_softmax: bool = False,
        need_head_weights: bool = False,
    ) -> Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:
        """
        Forward for MHA.

        Args:
            query (torch.Tensor): Query input.
            key (torch.Tensor, optional): Key input.
            value (torch.Tensor, optional): Value input.
            key_padding_mask (torch.Tensor, optional): Mask to exclude keys that
                are pads, of shape `(batch, src_len)`.
            need_weights (bool, optional): Return the attention weights,
                averaged overheads.
            attn_mask (torch.Tensor, optional): Typically used to implement
                causal attention.
            before_softmax (bool, optional): Return the raw attention weights and
                values before the attention softmax.
            need_head_weights (bool, optional): Return the attention weights for
                each head. Implies *need_weights*.
        """
        if need_head_weights:
            need_weights = True

        tgt_len, bsz, embed_dim = query.size()
        assert embed_dim == self.embed_dim
        assert list(query.size()) == [tgt_len, bsz, embed_dim]

        if self.self_attention:
            q = self.q_proj(query)
            k = self.k_proj(query)
            v = self.v_proj(query)
        elif self.encoder_decoder_attention:
            q = self.q_proj(query)
            if key is None:
                assert value is None
                k = v = None
            else:
                k = self.k_proj(key)
                v = self.v_proj(key)
        else:
            assert key is not None and value is not None
            q = self.q_proj(query)
            k = self.k_proj(key)
            v = self.v_proj(value)
        q *= self.scaling

        if self.bias_k is not None:
            assert self.bias_v is not None
            k = torch.cat([k, self.bias_k.repeat(1, bsz, 1)])
            v = torch.cat([v, self.bias_v.repeat(1, bsz, 1)])
            if attn_mask is not None:
                attn_mask = torch.cat(
                    [attn_mask, attn_mask.new_zeros(attn_mask.size(0), 1)], dim=1
                )
            if key_padding_mask is not None:
                key_padding_mask = torch.cat(
                    [
                        key_padding_mask,
                        key_padding_mask.new_zeros(key_padding_mask.size(0), 1),
                    ],
                    dim=1,
                )

        q = (
            q.contiguous()
            .view(tgt_len, bsz * self.num_heads, self.head_dim)
            .transpose(0, 1)
        )
        if k is not None:
            k = (
                k.contiguous()
                .view(-1, bsz * self.num_heads, self.head_dim)
                .transpose(0, 1)
            )
        if v is not None:
            v = (
                v.contiguous()
                .view(-1, bsz * self.num_heads, self.head_dim)
                .transpose(0, 1)
            )

        assert k is not None
        src_len = k.size(1)

        """
        This is part of a workaround to get around fork/join parallelism
        not supporting Optional types.
        """
        if key_padding_mask is not None and key_padding_mask.dim() == 0:
            key_padding_mask = None

        if key_padding_mask is not None:
            assert key_padding_mask.size(0) == bsz
            assert key_padding_mask.size(1) == src_len

        if self.add_zero_attn:
            assert v is not None
            src_len += 1
            k = torch.cat([k, k.new_zeros((k.size(0), 1) + k.size()[2:])], dim=1)
            v = torch.cat([v, v.new_zeros((v.size(0), 1) + v.size()[2:])], dim=1)
            if attn_mask is not None:
                attn_mask = torch.cat(
                    [attn_mask, attn_mask.new_zeros(attn_mask.size(0), 1)], dim=1
                )
            if key_padding_mask is not None:
                key_padding_mask = torch.cat(
                    [
                        key_padding_mask,
                        torch.zeros(key_padding_mask.size(0), 1).type_as(
                            key_padding_mask
                        ),
                    ],
                    dim=1,
                )

        attn_weights = torch.bmm(q, k.transpose(1, 2))

        assert list(attn_weights.size()) == [bsz * self.num_heads, tgt_len, src_len]

        if attn_mask is not None:
            attn_mask = attn_mask.unsqueeze(0)
            if self.onnx_trace:
                attn_mask = attn_mask.repeat(attn_weights.size(0), 1, 1)
            attn_weights += attn_mask

        if key_padding_mask is not None:
            """don't attend to padding symbols"""
            attn_weights = attn_weights.view(bsz, self.num_heads, tgt_len, src_len)
            if not self.tpu:
                attn_weights = attn_weights.masked_fill(
                    key_padding_mask.unsqueeze(1).unsqueeze(2).to(torch.bool),
                    float("-inf"),
                )
            else:
                attn_weights = attn_weights.transpose(0, 2)
                attn_weights = attn_weights.masked_fill(key_padding_mask, float("-inf"))
                attn_weights = attn_weights.transpose(0, 2)
            attn_weights = attn_weights.view(bsz * self.num_heads, tgt_len, src_len)

        if before_softmax:
            return attn_weights, v

        attn_weights_float = F.softmax(attn_weights, dim=-1)

        attn_weights = attn_weights_float.type_as(attn_weights)
        attn_probs = self.dropout_module(attn_weights)

        assert v is not None
        attn = torch.bmm(attn_probs, v)
        assert list(attn.size()) == [bsz * self.num_heads, tgt_len, self.head_dim]
        if self.onnx_trace and attn.size(1) == 1:
            """
            When ONNX tracing a single decoder step (sequence length == 1)
            the transpose is a no-op copy before view, thus unnecessary.
            """
            attn = attn.contiguous().view(tgt_len, bsz, embed_dim)
        else:
            attn = attn.transpose(0, 1).contiguous().view(tgt_len, bsz, embed_dim)
        attn = self.out_proj(attn)

        if need_weights:
            attn_weights = attn_weights_float.view(
                bsz, self.num_heads, tgt_len, src_len
            )
            if not need_head_weights:
                """average attention weights over heads"""
                attn_weights = attn_weights.mean(dim=1)

            return attn, attn_weights

        return attn


class SelfAttention2D(MultiHeadAttention):
    """
    COMPUTE SELF-ATTENTION OVER 2D INPUT

    Perform self-attention over 2D input for node features using multi-head
    attention.

    Args:
        embed_dim (int): Number of embedded dimensions for node features.
        num_heads (int): Number of heads for multi-head attention.
        axis (int): Indicate the axis over which the attention is performed.
        dropout (float): Dropout probability.
        max_size (int): Maximum size of the batch.
    """

    def __init__(
        self, embed_dim: int, num_heads: int, axis=None, dropout=0.0, max_size=4194304
    ):
        super(SelfAttention2D, self).__init__(
            embed_dim, num_heads, dropout=dropout, self_attention=True
        )
        self.axis = axis
        self.max_size = max_size

    def forward(self, x: torch.Tensor, padding_mask=None) -> torch.Tensor:
        """
        Forward self-attention over 2D-edge features.

        Reshape X depending on the axis attention mode!
        flatten over rows and cols for full N*M*N*M attention.

        Args:
            x (torch.Tensor): Edge feature self-attention update.
                [num_rows X num_cols X batch_size X embed_dim].
            padding_mask (torch.Tensor): Optional padding mask.
                [batch_size X num_rows X num_cols].
        """

        R, C, B, DIM = x.size()
        axis = self.axis
        if axis is None:
            x = x.view(R * C, B, DIM)
            if padding_mask is not None:
                padding_mask = padding_mask.view(B, R * C)
        else:
            assert axis == 0 or axis == 1

            """attend along the row dimension"""
            if axis == 0:
                x = x.view(R, C * B, DIM)
                if padding_mask is not None:
                    padding_mask = padding_mask.permute(2, 0, 1)
                    padding_mask = padding_mask.reshape(C * B, R)
                """attend along the col dimension"""
            else:
                x = x.transpose(0, 1)
                x = x.reshape(C, R * B, DIM)
                if padding_mask is not None:
                    padding_mask = padding_mask.permute(1, 0, 2)
                    padding_mask = padding_mask.reshape(R * B, C)

        if 0 < self.max_size < x.size(0) ** 2 * x.size(1):
            """
            Attention matrix size times batch size will exceed maximum
            allowable entries split into batches to make attention matrix RAM
            workable calculating attention over batches helps reduce RAM when
            N or M is large
            """
            batch_size = x.size(0) ** 2 // self.max_size
            if batch_size < 1:
                """might run out of RAM, but batch size can't be < 1"""
                batch_size = 1

            h = []
            for i in range(0, x.size(1), batch_size):
                xi = x[:, i : i + batch_size]
                mask = None
                if padding_mask is not None:
                    mask = padding_mask[i : i + batch_size]
                h.append(
                    super(SelfAttention2D, self).forward(xi, key_padding_mask=mask)
                )
            h = torch.cat(h, 1)
        else:
            h = super(SelfAttention2D, self).forward(x, key_padding_mask=padding_mask)

        """transpose h back to input shape"""
        if axis is None:
            h = h.view(R, C, B, DIM)
        elif axis == 0:
            h = h.view(R, C, B, DIM)
        else:
            h = h.view(C, R, B, DIM)
            h = h.transpose(0, 1)

        return h


class GeluFeedForward(nn.Module):
    """
    FEED-FORWARD TRANSFORMER MODULE USING GELU

    Input: Batch x ... x Dim
    Output: Batch x ... x Dim

    Args:
        input_dim (int): Number of input dimensions for linear transformation.
        ff_dim (int): Number of feed-forward dimensions in linear transformation.
    """

    def __init__(self, input_dim: int, ff_dim: int):
        super().__init__()
        self.norm = nn.LayerNorm(input_dim)
        self.linear1 = nn.Linear(input_dim, ff_dim)
        self.linear2 = nn.Linear(ff_dim, input_dim)

        nn.init.constant_(self.linear2.weight, 0.0)
        nn.init.constant_(self.linear2.bias, 0.0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward Gelu normalized tensor.

        Args:
            x (torch.Tensor): Any Tensor of shape [B x ... x D].

        Returns:
            torch.Tensor: Gelu normalized tensor.
        """
        x = self.norm(x)
        x = self.linear1(x)
        x = self.linear2(gelu(x))

        return x


@torch.jit.script
def gelu(x: torch.Tensor) -> torch.Tensor:
    """
    CUSTOM GAUSSIAN ERROR LINEAR UNITS ACTIVATION FUNCTION

    Args:
        x (torch.Tensor): Tensor input for activation.

    Returns:
        torch.Tensor: Gelu transform Tensor.
    """
    return x * 0.5 * (1.0 + torch.erf(x / sqrt(2)))



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
from os import getcwd
from os.path import join

import numpy as np
import torch
from torch import nn

from tardis_em.dist_pytorch.utils.segment_point_cloud import PropGreedyGraphCut
from tardis_em.utils.metrics import eval_graph_f1, mcov
from tardis_em.utils.trainer import BasicTrainer


class SparseDistTrainer(BasicTrainer):
    """
    DIST MODEL TRAINER
    """

    def __init__(self, **kwargs):
        super(SparseDistTrainer, self).__init__(**kwargs)

        self.node_input = self.structure["node_input"]

        self.Graph_gt = PropGreedyGraphCut(threshold=0.5, connection=1000)
        self.Graph0_25 = PropGreedyGraphCut(
            threshold=0.25, connection=self.instance_cov
        )
        self.Graph0_5 = PropGreedyGraphCut(threshold=0.5, connection=self.instance_cov)
        self.Graph0_9 = PropGreedyGraphCut(threshold=0.9, connection=self.instance_cov)

        self.mCov0_25, self.mCov0_5, self.mCov0_9 = [], [], []

    @staticmethod
    def _update_desc(stop_count: int, metric: list) -> str:
        """
        Utility function to update progress bar description.

        Args:
            stop_count (int): Early stop count.
            metric (list): Best f1 and mCov score.

        Returns:
            str: Updated progress bar status.
        """
        desc = (
            f"Epochs: early_stop: {stop_count}; "
            f"F1: [{metric[0]:.2f}; {metric[1]:.2f}]; "
            f"mCov 0.5: [{metric[2]:.2f}; {metric[3]:.2f}]; "
            f"mCov 0.9: [{metric[4]:.2f}; {metric[5]:.2f}]"
        )
        return desc

    def _update_epoch_desc(self):
        # For each Epoch load be t model from previous run
        if self.id == 0:
            self.epoch_desc = "Epochs: early_stop: 0; best F1: NaN"
        else:
            self.epoch_desc = self._update_desc(
                self.early_stopping.counter,
                [
                    np.max(self.f1) if len(self.f1) > 0 else 0.0,
                    self.f1[-1:][0] if len(self.f1) > 0 else 0.0,
                    np.max(self.mCov0_5) if len(self.mCov0_5) > 0 else 0.0,
                    self.mCov0_5[-1:][0] if len(self.mCov0_5) > 0 else 0.0,
                    np.max(self.mCov0_9) if len(self.mCov0_9) > 0 else 0.0,
                    self.mCov0_9[-1:][0] if len(self.mCov0_9) > 0 else 0.0,
                ],
            )

    def _save_metric(self) -> bool:
        """Save training metrics"""
        if len(self.training_loss) > 0:
            np.savetxt(
                join(
                    getcwd(),
                    f"{self.checkpoint_name}_checkpoint",
                    "training_losses.csv",
                ),
                self.training_loss,
                delimiter=",",
            )
        if len(self.validation_loss) > 0:
            np.savetxt(
                join(
                    getcwd(),
                    f"{self.checkpoint_name}_checkpoint",
                    "validation_losses.csv",
                ),
                self.validation_loss,
                delimiter=",",
            )
        if len(self.f1) > 0:
            np.savetxt(
                join(getcwd(), f"{self.checkpoint_name}_checkpoint", "eval_metric.csv"),
                np.column_stack(
                    [
                        self.accuracy,
                        self.precision,
                        self.recall,
                        self.threshold,
                        self.f1,
                        self.mCov0_25,
                        self.mCov0_5,
                        self.mCov0_9,
                    ]
                ),
                delimiter=",",
            )
        if len(self.learning_rate) > 0:
            np.savetxt(
                join(
                    getcwd(), f"{self.checkpoint_name}_checkpoint", "learning_rate.csv"
                ),
                self.learning_rate,
                delimiter=",",
            )

        """ Save current model weights"""
        # If mean evaluation f1 score is higher than save checkpoint
        if all(self.f1[-1:][0] >= i for i in self.f1[:-1]):
            torch.save(
                {
                    "model_struct_dict": self.structure,
                    "model_state_dict": self.model.state_dict(),
                    "optimizer_state_dict": self.optimizer.state_dict(),
                },
                join(
                    getcwd(),
                    f"{self.checkpoint_name}_checkpoint",
                    f"{self.checkpoint_name}_checkpoint_f1.pth",
                ),
            )

        # If mean evaluation mcov score is higher than save checkpoint
        if all(self.mCov0_9[-1:][0] >= i for i in self.mCov0_9[:-1]):
            torch.save(
                {
                    "model_struct_dict": self.structure,
                    "model_state_dict": self.model.state_dict(),
                    "optimizer_state_dict": self.optimizer.state_dict(),
                },
                join(
                    getcwd(),
                    f"{self.checkpoint_name}_checkpoint",
                    f"{self.checkpoint_name}_checkpoint_mcov.pth",
                ),
            )

        torch.save(
            {
                "model_struct_dict": self.structure,
                "model_state_dict": self.model.state_dict(),
                "optimizer_state_dict": self.optimizer.state_dict(),
            },
            join(getcwd(), f"{self.checkpoint_name}_checkpoint", "model_weights.pth"),
        )

        if self.early_stopping.early_stop:
            return True
        return False

    def _train(self):
        """
        Run model training.
        """
        # Update progress bar
        self._update_progress_bar(
            loss_desc="Training: (loss 1.000)", idx=0, task="Start Training..."
        )

        # Run training for DIST model
        for idx, (e, n, g, _, _) in enumerate(self.training_DataLoader):
            """Mid-training eval"""
            # self._update_progress_bar(
            #     loss_desc="Training: (loss 1.000)", idx=0, task="Mid-train Eval..."
            # )
            if len(self.training_DataLoader) > 100:
                self._mid_training_eval(idx=idx)

            """Training"""
            for edge, node, graph in zip(e, n, g):
                edge, graph = edge.to(self.device), graph.to(self.device)
                self.optimizer.zero_grad()

                edge, indices = self.model(coord=edge[0, :])
                graph = graph[0, indices[3][:, 0], indices[3][:, 1]].type(torch.float32)

                # Back-propagate
                loss = self.criterion(edge[:, 0], graph)
                loss.backward()  # One backward pass
                self.optimizer.step()  # Update the parameters

                # Store training loss metric
                loss_value = loss.item()
                self.training_loss.append(loss_value)

                # Store and update learning rate
                if self.lr_scheduler:
                    self.lr = self.optimizer.get_lr_scale()
                self.learning_rate.append(self.lr)

                # Update progress bar
                self._update_progress_bar(
                    loss_desc=f"Training: (loss {loss_value:.4f};"
                    f" LR: {self.lr:.5f})",
                    idx=idx,
                    task="Training...",
                )

    def _validate(self):
        """
        Test model against validation dataset.
        """
        valid_losses = []
        accuracy_mean = []
        precision_mean = []
        recall_mean = []
        F1_mean = []
        threshold_mean = []
        mcov0_25, mcov0_5, mcov0_9 = [], [], []

        for idx, (e, n, g, o, _) in enumerate(self.validation_DataLoader):
            coord = [x.cpu().detach().numpy()[0, :] for x in e]
            edge_cpu, graph_cpu, out_cpu = [], [], []

            for edge, node, graph, out in zip(e, n, g, o):
                edge, graph = edge[0, :].to(self.device), graph.to(self.device)

                with torch.no_grad():
                    out_cpu.append(out.cpu().detach().numpy()[0, :])

                    # Predict graph
                    edge, indices = self.model(coord=edge)

                    # Calculate late validation loss
                    loss = self.criterion(
                        edge[:, 0],
                        graph[0, indices[3][:, 0], indices[3][:, 1]].type(
                            torch.float32
                        ),
                    )

                    # Calculate F1 metric
                    pred_edge = np.zeros(indices[2])
                    pred_edge[indices[3][:, 0], indices[3][:, 1]] += (
                        edge.cpu().detach().numpy()[:, 0]
                    )
                    np.fill_diagonal(pred_edge, 1)
                    edge_cpu.append(pred_edge)
                    graph_cpu.append(graph[0, :].cpu().detach().numpy())

                    acc, prec, recall, f1, th = eval_graph_f1(
                        logits=torch.from_numpy(pred_edge),
                        targets=graph[0, ...].cpu().detach(),
                        threshold=0.5,
                    )

                # Avg. precision score
                valid_losses.append(loss.item())
                accuracy_mean.append(acc)
                precision_mean.append(prec)
                recall_mean.append(recall)
                F1_mean.append(f1)
                threshold_mean.append(th)

                # Update progress bar
                df_05, df_09 = 0.0, 0.0
                if len(mcov0_5) > 0:
                    df_05 = mcov0_5[-1:][0]
                if len(mcov0_9) > 0:
                    df_09 = mcov0_9[-1:][0]

                valid = (
                    f"Validation: (loss: {loss.item():.4f}; F1: {f1:.2f}) "
                    f"mCov[0.5]: {df_05:.2f}; "
                    f"mCov[0.9]: {df_09:.2f}"
                )
                self._update_progress_bar(loss_desc=valid, idx=idx, train=False)

            # Build GT instance point cloud
            target = self.Graph_gt.patch_to_segment(
                graph=graph_cpu, coord=coord, idx=out_cpu, prune=0, sort=False
            )

            # Threshold 0.25
            try:
                input0_1 = self.Graph0_25.patch_to_segment(
                    graph=edge_cpu, coord=coord, idx=out_cpu, prune=5, sort=False
                )
                mcov_m, _ = mcov(input0_1, target)
                mcov0_25.append(mcov_m)
            except:
                mcov0_25.append(0.0)

            # Threshold 0.5
            try:
                input0_5 = self.Graph0_5.patch_to_segment(
                    graph=edge_cpu, coord=coord, idx=out_cpu, prune=5, sort=False
                )
                mcov_m, _ = mcov(input0_5, target)
                mcov0_5.append(mcov_m)
            except:
                mcov0_5.append(0.0)

            # Threshold 0.9
            try:
                input0_9 = self.Graph0_9.patch_to_segment(
                    graph=edge_cpu, coord=coord, idx=out_cpu, prune=5, sort=False
                )
                mcov_m, _ = mcov(input0_9, target)
                mcov0_9.append(mcov_m)
            except:
                mcov0_9.append(0.0)

            # Update progress bar
            valid = (
                f"Validation: (loss: {loss.item():.4f}; F1: {f1:.2f}) "
                f"mCov[0.5]: {mcov0_5[-1:][0]:.2f}; "
                f"mCov[0.9]: {mcov0_9[-1:][0]:.2f}"
            )
            self._update_progress_bar(loss_desc=valid, idx=idx, train=False)

        # Reduce eval. metric with mean
        self.validation_loss.append(np.mean(valid_losses))
        self.accuracy.append(np.mean(accuracy_mean))
        self.precision.append(np.mean(precision_mean))
        self.recall.append(np.mean(recall_mean))
        self.threshold.append(np.mean(threshold_mean))
        self.f1.append(np.mean(F1_mean))

        self.mCov0_25.append(np.mean(mcov0_25))
        self.mCov0_5.append(np.mean(mcov0_5))
        self.mCov0_9.append(np.mean(mcov0_9))

        # Check if average evaluation loss dropped
        self.early_stopping(f1_score=self.f1[-1:][0])


class DistTrainer(BasicTrainer):
    """
    DIST MODEL TRAINER
    """

    def __init__(self, **kwargs):
        super(DistTrainer, self).__init__(**kwargs)

        self.node_input = self.structure["node_input"]

        self.Graph_gt = PropGreedyGraphCut(threshold=0.5, connection=1000)
        self.Graph0_25 = PropGreedyGraphCut(
            threshold=0.25, connection=self.instance_cov
        )
        self.Graph0_5 = PropGreedyGraphCut(threshold=0.5, connection=self.instance_cov)
        self.Graph0_9 = PropGreedyGraphCut(threshold=0.9, connection=self.instance_cov)

        self.mCov0_25, self.mCov0_5, self.mCov0_9 = [], [], []

    @staticmethod
    def _update_desc(stop_count: int, metric: list) -> str:
        """
        Utility function to update progress bar description.

        Args:
            stop_count (int): Early stop count.
            metric (list): Best f1 and mCov score.

        Returns:
            str: Updated progress bar status.
        """
        desc = (
            f"Epochs: early_stop: {stop_count}; "
            f"F1: [{metric[0]:.2f}; {metric[1]:.2f}]; "
            f"mCov 0.5: [{metric[2]:.2f}; {metric[3]:.2f}]; "
            f"mCov 0.9: [{metric[4]:.2f}; {metric[5]:.2f}]"
        )
        return desc

    def _update_epoch_desc(self):
        # For each Epoch load be t model from previous run
        if self.id == 0:
            self.epoch_desc = "Epochs: early_stop: 0; best F1: NaN"
        else:
            self.epoch_desc = self._update_desc(
                self.early_stopping.counter,
                [
                    np.max(self.f1) if len(self.f1) > 0 else 0.0,
                    self.f1[-1:][0] if len(self.f1) > 0 else 0.0,
                    np.max(self.mCov0_5) if len(self.mCov0_5) > 0 else 0.0,
                    self.mCov0_5[-1:][0] if len(self.mCov0_5) > 0 else 0.0,
                    np.max(self.mCov0_9) if len(self.mCov0_9) > 0 else 0.0,
                    self.mCov0_9[-1:][0] if len(self.mCov0_9) > 0 else 0.0,
                ],
            )

    def _save_metric(self) -> bool:
        """Save training metrics"""
        if len(self.training_loss) > 0:
            np.savetxt(
                join(
                    getcwd(),
                    f"{self.checkpoint_name}_checkpoint",
                    "training_losses.csv",
                ),
                self.training_loss,
                delimiter=",",
            )
        if len(self.validation_loss) > 0:
            np.savetxt(
                join(
                    getcwd(),
                    f"{self.checkpoint_name}_checkpoint",
                    "validation_losses.csv",
                ),
                self.validation_loss,
                delimiter=",",
            )
        if len(self.f1) > 0:
            np.savetxt(
                join(getcwd(), f"{self.checkpoint_name}_checkpoint", "eval_metric.csv"),
                np.column_stack(
                    [
                        self.accuracy,
                        self.precision,
                        self.recall,
                        self.threshold,
                        self.f1,
                        self.mCov0_25,
                        self.mCov0_5,
                        self.mCov0_9,
                    ]
                ),
                delimiter=",",
            )
        if len(self.learning_rate) > 0:
            np.savetxt(
                join(
                    getcwd(), f"{self.checkpoint_name}_checkpoint", "learning_rate.csv"
                ),
                self.learning_rate,
                delimiter=",",
            )

        """ Save current model weights"""
        # If mean evaluation f1 score is higher than save checkpoint
        if all(self.f1[-1:][0] >= i for i in self.f1[:-1]):
            torch.save(
                {
                    "model_struct_dict": self.structure,
                    "model_state_dict": self.model.state_dict(),
                    "optimizer_state_dict": self.optimizer.state_dict(),
                },
                join(
                    getcwd(),
                    f"{self.checkpoint_name}_checkpoint",
                    f"{self.checkpoint_name}_checkpoint_f1.pth",
                ),
            )

        # If mean evaluation mcov score is higher than save checkpoint
        if all(self.mCov0_9[-1:][0] >= i for i in self.mCov0_9[:-1]):
            torch.save(
                {
                    "model_struct_dict": self.structure,
                    "model_state_dict": self.model.state_dict(),
                    "optimizer_state_dict": self.optimizer.state_dict(),
                },
                join(
                    getcwd(),
                    f"{self.checkpoint_name}_checkpoint",
                    f"{self.checkpoint_name}_checkpoint_mcov.pth",
                ),
            )

        torch.save(
            {
                "model_struct_dict": self.structure,
                "model_state_dict": self.model.state_dict(),
                "optimizer_state_dict": self.optimizer.state_dict(),
            },
            join(getcwd(), f"{self.checkpoint_name}_checkpoint", "model_weights.pth"),
        )

        if self.early_stopping.early_stop:
            return True
        return False

    def _train(self):
        """
        Run model training.
        """
        # Update progress bar
        self._update_progress_bar(
            loss_desc="Training: (loss 1.000)", idx=0, task="Start Training..."
        )

        # Run training for DIST model
        for idx, (e, n, g, _, _) in enumerate(self.training_DataLoader):
            """Mid-training eval"""
            # self._update_progress_bar(
            #     loss_desc="Training: (loss 1.000)", idx=0, task="Mid-train Eval..."
            # )
            if len(self.training_DataLoader) > 100:
                self._mid_training_eval(idx=idx)

            """Training"""
            for edge, node, graph in zip(e, n, g):
                edge, graph = edge.to(self.device), graph.to(self.device)
                self.optimizer.zero_grad()

                if self.node_input > 0:
                    edge = self.model(coords=edge, node_features=node.to(self.device))
                else:
                    edge = self.model(coords=edge, node_features=None)

                # Back-propagate
                loss = self.criterion(edge[:, 0, ...], graph)  # Calc. loss
                loss.backward()  # One backward pass
                self.optimizer.step()  # Update the parameters

                # Store training loss metric
                loss_value = loss.item()
                self.training_loss.append(loss_value)

                # Store and update learning rate
                if self.lr_scheduler:
                    self.lr = self.optimizer.get_lr_scale()
                self.learning_rate.append(self.lr)

                # Update progress bar
                self._update_progress_bar(
                    loss_desc=f"Training: (loss {loss_value:.4f};"
                    f" LR: {self.lr:.5f})",
                    idx=idx,
                    task="Training...",
                )

    @staticmethod
    def _greedy_segmenter(graph: np.ndarray, coord: np.ndarray, th: float):
        node = dict()
        for j in range(len(graph)):
            df = [[id_, i] for id_, i in enumerate(graph[j]) if i >= th and id != j]
            node[j] = dict(df)
        coord_list = dict()
        seg_id = 0
        full_list = list(range(len(graph)))

        while len(full_list) != 0:
            id_, old_id = [full_list[0]], [full_list[0]]
            added = list(node[id_[0]].keys())
            id_ = id_ + added

            while len(id_) != len(old_id):
                old_id = id_
                added_new = []
                for key in added:
                    added_new = added_new + list(node[key].keys())
                added = list(dict.fromkeys(added_new))

                id_ = list(dict.fromkeys(id_ + added))
            id_ = sorted(id_)

            coord_list[seg_id] = id_
            seg_id += 1

            for i in id_:
                full_list.remove(i)

        list_c = []
        for key in coord_list:
            value = coord[coord_list[key], :]
            id_ = np.repeat(key, len(value)).reshape(-1, 1)
            list_c.append(np.hstack((id_, value)))
        return np.concatenate(list_c)

    def _validate(self):
        """
        Test model against validation dataset.
        """
        valid_losses = []
        accuracy_mean = []
        precision_mean = []
        recall_mean = []
        F1_mean = []
        threshold_mean = []
        mcov0_25, mcov0_5, mcov0_9 = [], [], []

        for idx, (e, n, g, o, _) in enumerate(self.validation_DataLoader):
            coord = [x.cpu().detach().numpy()[0, :] for x in e]
            edge_cpu, graph_cpu, out_cpu = [], [], []

            for edge, node, graph, out in zip(e, n, g, o):
                edge, graph = edge.to(self.device), graph.to(self.device)

                with torch.no_grad():
                    out_cpu.append(out.cpu().detach().numpy()[0, :])
                    graph_cpu.append(graph[0, :].cpu().detach().numpy())

                    # Predict graph
                    if self.node_input > 0:
                        edge = self.model(
                            coords=edge, node_features=node.to(self.device)
                        )
                    else:
                        edge = self.model(coords=edge, node_features=None)

                    # Calculate validation loss
                    loss = self.criterion(edge[0, :], graph)

                    # Calculate F1 metric
                    edge[0, 0, :].fill_diagonal_(1)
                    acc, prec, recall, f1, th = eval_graph_f1(
                        logits=edge[0, 0, :], targets=graph[0, :], threshold=0.5
                    )
                    edge_cpu.append(edge[0, 0, :].cpu().detach().numpy())

                # Avg. precision score
                valid_losses.append(loss.item())
                accuracy_mean.append(acc)
                precision_mean.append(prec)
                recall_mean.append(recall)
                F1_mean.append(f1)
                threshold_mean.append(th)

                # Update progress bar
                df_05, df_09 = 0.0, 0.0
                if len(mcov0_5) > 0:
                    df_05 = mcov0_5[-1:][0]
                if len(mcov0_9) > 0:
                    df_09 = mcov0_9[-1:][0]

                valid = (
                    f"Validation: (loss: {loss.item():.4f}; F1: {f1:.2f}) "
                    f"mCov[0.5]: {df_05:.2f}; "
                    f"mCov[0.9]: {df_09:.2f}"
                )
                self._update_progress_bar(loss_desc=valid, idx=idx, train=False)

            # Build GT instance point cloud
            target = self.Graph_gt.patch_to_segment(
                graph=graph_cpu, coord=coord, idx=out_cpu, prune=0, sort=False
            )

            # Threshold 0.25
            try:
                input0_1 = self.Graph0_25.patch_to_segment(
                    graph=edge_cpu, coord=coord, idx=out_cpu, prune=5, sort=False
                )
                mcov_m, _ = mcov(input0_1, target)
                mcov0_25.append(mcov_m)
            except:
                mcov0_25.append(0.0)

            # Threshold 0.5
            try:
                input0_5 = self.Graph0_5.patch_to_segment(
                    graph=edge_cpu, coord=coord, idx=out_cpu, prune=5, sort=False
                )
                mcov_m, _ = mcov(input0_5, target)
                mcov0_5.append(mcov_m)
            except:
                mcov0_5.append(0.0)

            # Threshold 0.9
            try:
                input0_9 = self.Graph0_9.patch_to_segment(
                    graph=edge_cpu, coord=coord, idx=out_cpu, prune=5, sort=False
                )
                mcov_m, _ = mcov(input0_9, target)
                mcov0_9.append(mcov_m)
            except:
                mcov0_9.append(0.0)

            # Update progress bar
            valid = (
                f"Validation: (loss: {loss.item():.4f}; F1: {f1:.2f}) "
                f"mCov[0.5]: {mcov0_5[-1:][0]:.2f}; "
                f"mCov[0.9]: {mcov0_9[-1:][0]:.2f}"
            )
            self._update_progress_bar(loss_desc=valid, idx=idx, train=False)

        # Reduce eval. metric with mean
        self.validation_loss.append(np.mean(valid_losses))
        self.accuracy.append(np.mean(accuracy_mean))
        self.precision.append(np.mean(precision_mean))
        self.recall.append(np.mean(recall_mean))
        self.threshold.append(np.mean(threshold_mean))
        self.f1.append(np.mean(F1_mean))

        self.mCov0_25.append(np.mean(mcov0_25))
        self.mCov0_5.append(np.mean(mcov0_5))
        self.mCov0_9.append(np.mean(mcov0_9))

        # Check if average evaluation loss dropped
        self.early_stopping(f1_score=self.f1[-1:][0])


class CDistTrainer(BasicTrainer):
    """
    C_DIST MODEL TRAINER
    """

    def __init__(self, **kwargs):
        super(CDistTrainer, self).__init__(**kwargs)

        if self.structure["dist_type"] == "semantic":
            self.criterion_cls = nn.CrossEntropyLoss()

    def _train(self):
        """
        Run model training.
        """
        # Update progress bar
        self._update_progress_bar(loss_desc="Training: (loss 1.000)", idx=0)

        for idx, (e, n, g, _, c) in enumerate(self.training_DataLoader):
            """Mid-training eval"""
            self._mid_training_eval(idx=idx)

            for edge, node, cls, graph in zip(e, n, c, g):
                edge, graph = edge.to(self.device), graph.to(self.device)
                cls = cls.to(self.device)
                self.optimizer.zero_grad()

                if self.node_input:
                    node = node.to(self.device)
                    edge, out_cls = self.model(coords=edge, node_features=node)
                else:
                    edge, out_cls = self.model(coords=edge, node_features=None)

                # Back-propagate
                loss = self.criterion(edge[0, :], graph) + self.criterion_cls(
                    out_cls, cls
                )
                loss.backward()  # One backward pass
                self.optimizer.step()  # Update the parameters

                # Store evaluation loss metric
                loss_value = loss.item()
                self.training_loss.append(loss_value)

                # Update progress bar
                self._update_progress_bar(
                    loss_desc=f"Training: (loss {loss_value:.4f})", idx=idx
                )

    def _validate(self):
        """
        Test model against validation dataset.
        """
        valid_losses = []
        accuracy_mean = []
        precision_mean = []
        recall_mean = []
        F1_mean = []
        threshold_mean = []

        for idx, (e, n, g, _, c) in enumerate(self.validation_DataLoader):
            for edge, node, cls, graph in zip(e, n, c, g):
                edge, graph = edge.to(self.device), graph.to(self.device)
                cls = cls.to(self.device)

                with torch.no_grad():
                    if self.node_input:
                        node = node.to(self.device)
                        edge, out_cls = self.model(coords=edge, node_features=node)
                    else:
                        edge, out_cls = self.model(coords=edge, node_features=None)

                    loss = self.criterion(edge[0, :], graph) + self.criterion_cls(
                        out_cls, cls
                    )

                    edge = torch.sigmoid(edge[:, 0, :])
                    acc, prec, recall, f1, th = eval_graph_f1(
                        logits=edge, targets=graph, threshold=0.5
                    )

                # Avg. precision score
                valid_losses.append(loss.item())
                accuracy_mean.append(acc)
                precision_mean.append(prec)
                recall_mean.append(recall)
                F1_mean.append(f1)
                threshold_mean.append(th)
                valid = (
                    "Validation: "
                    f"(loss {loss.item():.4f} Prec: {prec:.2f} Rec: {recall:.2f} F1: {f1:.2f})"
                )

                # Update progress bar
                self._update_progress_bar(
                    loss_desc=valid, idx=idx, task="Validation..."
                )
        # Reduce eval. metric with mean
        self.validation_loss.append(np.mean(valid_losses))
        self.accuracy.append(np.mean(accuracy_mean))
        self.precision.append(np.mean(precision_mean))
        self.recall.append(np.mean(recall_mean))
        self.threshold.append(np.mean(threshold_mean))
        self.f1.append(np.mean(F1_mean))

        # Check if average evaluation loss dropped
        self.early_stopping(f1_score=np.mean(F1_mean))



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
from abc import abstractmethod

import torch
import torch.nn as nn
import torch.nn.functional as F


class AbstractLoss(nn.Module):
    def __init__(self, smooth=1e-16, reduction="mean", diagonal=False, sigmoid=True):
        """
        Initializes the abstract loss function with the given parameters.

        Args:
            smooth (float): Smoothing factor to ensure no division by 0.
            reduction (str): The reduction to apply to the output: 'none' | 'mean' | 'sum'.
                'none': no reduction will be applied.
                'mean': output would be a mean values.
                'sum': the output will be summed.
            diagonal (bool): If True, the diagonal of the adjacency matrix
                is removed in graph predictions.
            sigmoid (bool): If True, compute sigmoid before loss.
        """
        super(AbstractLoss, self).__init__()
        self.smooth = smooth
        self.diagonal = diagonal
        self.sigmoid = sigmoid
        self.reduction = reduction

        assert self.reduction in ["sum", "mean", "none"]

    def activate(self, logits):
        if self.sigmoid:
            return torch.sigmoid(logits)
        return logits

    def ignor_diagonal(self, logits, targets, mask=False):
        if self.diagonal:
            g_range = range(logits.shape[-1])

            if mask:
                mask = torch.ones_like(targets)
                mask[:, g_range, g_range] = 0.0

                logits = logits * mask
                targets = targets * mask
            else:
                logits[:, g_range, g_range] = 1
                targets[:, g_range, g_range] = 1
            return logits, targets
        return logits, targets

    def initialize_tensors(self, logits, targets, mask):
        # Activation and optionally ignore diagonal for graphs
        logits = self.activate(logits)

        return self.ignor_diagonal(logits=logits, targets=targets, mask=mask)

    @abstractmethod
    def forward(self, logits: torch.Tensor, targets: torch.Tensor, mask=False):
        """

        Args:
            logits (torch.Tensor): The predicted logits. Shape: [Batch x Channels x ...].
            targets (torch.Tensor): The target values. Shape: [Batch x Channels x ...].
            mask (bool): If Ture, output mask diagonal axis.

        Return:
            torch.Tensor: Computed loss function.
        """
        pass


class AdaptiveDiceLoss(AbstractLoss):
    """
    Implements an adaptive Dice loss function, which gives more weight to false negatives.

    The AdaptiveDiceLoss loss function is a variant of the standard Dice loss,
    with an additional adaptive term (1 - logits) ** self.alpha applied to logits.
    This term will give higher weight to false negatives (i.e., the cases where
    the prediction is low but the ground truth is high),
    which can be useful in cases where these are particularly costly.
    """

    def __init__(self, alpha=0.1, **kwargs):
        """
        Initializes the AdaptiveDiceLoss with the given parameters.

        Args:
            alpha (float): The scaling factor for the adaptive term.
                Higher values give more weight to false negatives.
            smooth (float): A small constant to avoid division by zero.
        """
        super(AdaptiveDiceLoss, self).__init__(**kwargs)
        self.alpha = alpha

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor, mask=False
    ) -> torch.Tensor:
        """
        Computes the adaptive Dice loss between the logits and targets.
        """
        logits, targets = self.initialize_tensors(logits, targets, mask)

        # Soft weighted dice
        logits = logits.view(-1)
        targets = targets.view(-1)

        logits = ((1 - logits) ** self.alpha) * logits

        intersection = (logits * targets).sum()
        dice = (2.0 * intersection + self.smooth) / (
            logits.square().sum() + targets.square().sum() + self.smooth
        )

        return 1 - dice


class BCELoss(AbstractLoss):
    """
    Implements the Binary Cross-Entropy loss function with an option to ignore
    the diagonal elements.

    The BCELoss class can be used for training where pixel-level accuracy is important.
    """

    def __init__(self, **kwargs):
        """
        Initializes the BCELoss with the given parameters.
        """
        super(BCELoss, self).__init__(**kwargs)

        self.loss = nn.BCELoss(reduction=self.reduction)

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor, mask=True
    ) -> torch.Tensor:
        """
        Computes the BCE loss between the logits and targets.
        """
        logits, targets = self.initialize_tensors(logits, targets, mask)

        return self.loss(logits, targets)


class BCEGraphWiseLoss(AbstractLoss):
    """
    Implements the Binary Cross-Entropy loss function with an option to ignore
    the diagonal elements.

    The BCELoss class can be used for training where pixel-level accuracy is important.
    """

    def __init__(self, **kwargs):
        """
        Initializes the BCELoss with the given parameters.
        """
        super(BCEGraphWiseLoss, self).__init__(**kwargs)

        self.loss = nn.BCELoss(reduction=self.reduction)

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor, mask=True
    ) -> torch.Tensor:
        """
        Computes the BCE loss between the logits and targets.
        """
        logits, targets = self.initialize_tensors(logits, targets, mask)
        idx_1 = torch.where(targets > 0)
        idx_0 = torch.where(targets == 0)

        pos_loss = self.loss(logits[idx_1], targets[idx_1])
        neg_loss = self.loss(logits[idx_0], targets[idx_0])
        return pos_loss + neg_loss


class BCEDiceLoss(AbstractLoss):
    """
    DICE + BCE LOSS FUNCTION
    """

    def __init__(self, **kwargs):
        """
        Loss initialization
        """
        super(BCEDiceLoss, self).__init__(**kwargs)
        self.bce = BCELoss(diagonal=self.diagonal, sigmoid=self.sigmoid)
        self.dice = DiceLoss(diagonal=self.diagonal, sigmoid=self.sigmoid)

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor, mask=False
    ) -> torch.Tensor:
        """
        Forward loos function
        """
        bce_loss = self.bce(logits, targets, mask)
        dice_loss = self.dice(logits, targets, mask)

        if dice_loss is None:
            return bce_loss + 1

        return bce_loss + dice_loss


class CELoss(AbstractLoss):
    """
    STANDARD CROSS-ENTROPY LOSS FUNCTION
    """

    def __init__(self, **kwargs):
        """
        Loss initialization
        """
        super(CELoss, self).__init__(**kwargs)

        self.loss = nn.CrossEntropyLoss(reduction=self.reduction)

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor, mask=False
    ) -> torch.Tensor:
        """
        Forward loos function
        """
        logits, targets = self.initialize_tensors(logits, targets, mask)

        return self.loss(logits, targets)


class DiceLoss(AbstractLoss):
    """
    Dice coefficient loss function.

    Dice=2(A∩B)(A)+(B);
    where 'A∩B' represents the common elements between sets A and B
    'A' ann 'B' represents the number of elements in set A ans set B

    This loss effectively zero-out any pixels from our prediction which
    are not "activated" in the target mask.
    """

    def __init__(self, **kwargs):
        """
        Loss initialization

        Args:
            smooth (float): Smooth factor to ensure  no 0 division.
        """
        super(DiceLoss, self).__init__(**kwargs)

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor, mask=False
    ) -> torch.Tensor:
        """
        Forward loos function
        """
        logits, targets = self.initialize_tensors(logits, targets, mask)

        # Flatten label and prediction tensors
        logits = logits[0, :].view(-1)
        targets = targets[0, :].view(-1)

        # Calculate dice loss
        intersection = (logits * targets).sum()
        dice = (2 * intersection + self.smooth) / (
            logits.square().sum() + targets.square().sum() + self.smooth
        )

        return 1 - dice


class LaplacianEigenmapsLoss(AbstractLoss):
    """
    A loss function for deep learning models that computes the mean squared error
    between the first non-zero eigenvectors of the Laplacian matrices of the
    ground truth and predicted adjacency matrices.
    """

    def __init__(self, **kwargs):
        """
        Initializes the LaplacianEigenmapsLoss with an instance of nn.MSELoss
        """
        super(LaplacianEigenmapsLoss, self).__init__(**kwargs)
        self.mse_loss = nn.MSELoss(reduction=self.reduction)

    @staticmethod
    def compute_laplacian(A: torch.Tensor) -> torch.Tensor:
        """
        Computes the Laplacian matrix of an adjacency matrix.

        Args:
            A (torch.Tensor): The adjacency matrix.

        Returns:
            The Laplacian matrix.
        """
        D = torch.diag(torch.sum(A, dim=1))

        return D - A

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor, mask=False
    ) -> torch.Tensor:
        """
        Computes the Laplacian-Eigenmaps loss between the true and predicted adjacency matrices.
        """
        logits, targets = self.initialize_tensors(logits, targets, mask)

        # Compute Laplacian matrices
        L_true = self.compute_laplacian(logits)
        L_pred = self.compute_laplacian(targets)

        # Computes the smallest non-zero eigenvector of each Laplacian
        _, eigenvectors_true = torch.linalg.eigh(L_true)
        _, eigenvectors_pred = torch.linalg.eigh(L_pred)

        v_true = eigenvectors_true[:, 1]
        v_pred = eigenvectors_pred[:, 1]

        # computes the mean squared error
        return self.mse_loss(v_true, v_pred)


class SoftSkeletonization(AbstractLoss):
    """
    General soft skeletonization with DICE loss function
    """

    def __init__(self, _iter=5, **kwargs):
        """
        Loss initialization

        Args:
            _iter (int): Number of skeletonization iterations
        """
        super(SoftSkeletonization, self).__init__(**kwargs)

        self.iter = _iter

    @staticmethod
    def _soft_erode(binary_mask):
        if len(binary_mask.shape) == 4:  # 2D
            p1 = -F.max_pool2d(-binary_mask, (3, 1), (1, 1), (1, 0))
            p2 = -F.max_pool2d(-binary_mask, (1, 3), (1, 1), (0, 1))

            return torch.min(p1, p2)
        elif len(binary_mask.shape) == 5:  # 3D
            p1 = -F.max_pool3d(-binary_mask, (3, 1, 1), (1, 1, 1), (1, 0, 0))
            p2 = -F.max_pool3d(-binary_mask, (1, 3, 1), (1, 1, 1), (0, 1, 0))
            p3 = -F.max_pool3d(-binary_mask, (1, 1, 3), (1, 1, 1), (0, 0, 1))
            return torch.min(torch.min(p1, p2), p3)

    @staticmethod
    def _soft_dilate(binary_mask):
        if len(binary_mask.shape) == 4:  # 2D
            return F.max_pool2d(binary_mask, (3, 3), (1, 1), (1, 1))
        elif len(binary_mask.shape) == 5:  # 3D
            return F.max_pool3d(binary_mask, (3, 3, 3), (1, 1, 1), (1, 1, 1))

    def _soft_open(self, binary_mask):
        return self._soft_dilate(self._soft_erode(binary_mask))

    def soft_skel(self, binary_mask: torch.Tensor, iter_: int) -> torch.Tensor:
        """
        Soft skeletonization

        Args:
            binary_mask: Binary target mask.
            iter_: Number of iterations for erosion.

        Returns:
            torch.Tensor: Skeleton on-hot mask
        """
        if not isinstance(iter_, int):
            iter_ = int(iter_)

        binary_mask_open = self._soft_open(binary_mask)
        s_skeleton = F.relu(binary_mask - binary_mask_open)

        for j in range(iter_):
            binary_mask = self._soft_erode(binary_mask)
            binary_mask_open = self._soft_open(binary_mask)

            delta = F.relu(binary_mask - binary_mask_open)
            s_skeleton = s_skeleton + F.relu(delta - (s_skeleton * delta))

        return s_skeleton

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor, mask=False
    ) -> torch.Tensor:
        """
        Forward loos function
        """
        pass


class ClBCELoss(SoftSkeletonization):
    """
    Soft skeletonization with BCE loss function

    Implements a custom version of the Binary Cross Entropy (BCE) loss,
    where an additional term is added to the standard BCE loss.
    This additional term is a kind of F1 score calculated on the soft-skeletonized version
    of the predicted and target masks.
    """

    def __init__(self, **kwargs):
        super(ClBCELoss, self).__init__(**kwargs)
        self.bce = BCELoss(diagonal=self.diagonal, sigmoid=self.sigmoid)

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor, mask=False
    ) -> torch.Tensor:
        """
        Forward loss function
        """
        # BCE with activation
        bce = self.bce(logits, targets)

        # Activation and optionally omit diagonal
        logits, targets = self.initialize_tensors(logits, targets, mask)

        # Soft skeletonization
        sk_logits = self.soft_skel(logits, self.iter)
        sk_targets = self.soft_skel(targets, self.iter)

        t_prec = (sk_logits * targets).sum() + self.smooth
        t_prec /= sk_logits.sum() + self.smooth

        t_sens = (sk_targets * logits).sum() + self.smooth
        t_sens /= sk_targets.sum() + self.smooth

        # ClBCE
        cl_bce = 2 * (t_prec * t_sens) / (t_prec + t_sens)

        return bce + (1 - cl_bce)


class ClDiceLoss(SoftSkeletonization):
    """
    Soft skeletonization with DICE loss function

    Implements a custom version of the Dice loss,
    where an additional term is added to the standard BCE loss.
    This additional term is a kind of F1 score calculated on the soft-skeletonized version
    of the predicted and target masks.
    """

    def __init__(self, **kwargs):
        super(ClDiceLoss, self).__init__(**kwargs)
        self.soft_dice = DiceLoss(diagonal=self.diagonal, sigmoid=self.sigmoid)

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor, mask=False
    ) -> torch.Tensor:
        """
        Forward loss function
        """
        # Dice loss with activation
        dice = self.soft_dice(logits, targets)

        logits, targets = self.initialize_tensors(logits, targets, mask)
        # Soft skeletonization
        sk_logits = self.soft_skel(logits, self.iter)
        sk_targets = self.soft_skel(targets, self.iter)

        t_prec = (sk_logits * targets).sum() + self.smooth
        t_prec /= sk_logits.sum() + self.smooth

        t_sens = (sk_targets * logits).sum() + self.smooth
        t_sens /= sk_targets.sum() + self.smooth

        # CLDice loss
        cl_dice = 2 * (t_prec * t_sens) / (t_prec + t_sens)

        return dice + (1 - cl_dice)


class SigmoidFocalLoss(AbstractLoss):
    """
    Implements the Sigmoid Focal Loss function with an option to ignore the diagonal elements.

    The SigmoidFocalLoss class implements the Focal Loss, which was proposed
    as a method for focusing the model on hard examples during the training of an object detector.
    It provides an option to ignore the diagonal elements of the input matrices.

    References: 10.1088/1742-6596/1229/1/012045
    """

    def __init__(self, gamma=0.25, alpha=None, **kwargs):
        """
        Initializes the SigmoidFocalLoss with the given parameters.

        Args:
            gamma (float): The gamma factor used in the focal loss computation.
            alpha (float, optional): Optional alpha factor for class balance.
                If not provided, no class balance is performed.
        """
        super(SigmoidFocalLoss, self).__init__(**kwargs)
        self.gamma = gamma
        self.alpha = alpha

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor, mask=False
    ) -> torch.Tensor:
        """
        Computes the sigmoid focal loss between the logits and targets.
        """
        logits, targets = self.initialize_tensors(logits, targets, mask)

        # Compute focal loss term_1 and term_2
        term1 = (1 - logits) ** self.gamma * torch.log(logits)
        term1 = torch.where(torch.isinf(term1), 0, term1)

        term2 = logits**self.gamma * torch.log(1 - logits)
        term2 = torch.where(torch.isinf(term2), 0, term2)

        # Compute sigmoid focal loss
        y = targets.unsqueeze(1)
        if self.alpha is None:
            sfl = torch.mean(y * term1 + ((1 - y) * term2))
            sfl = -(1 / logits.size()[0]) * sfl
        else:
            sfl = torch.mean(
                self.alpha * y * term1 + (1 - self.alpha) * (1 - y) * term2
            )

        return sfl


class WBCELoss(AbstractLoss):
    """
    Implements a weighted Binary Cross-Entropy loss function with an option
    to ignore the diagonal elements.

    The WBCELoss class can help to balance the contribution of positive and
    negative samples in datasets
    where one class significantly outnumbers the other.
    It provides an option to ignore the diagonal elements of the input matrices,
    which could be useful for applications like graph prediction
    where self-connections might not be meaningful.
    """

    def __init__(self, **kwargs):
        """
        Initializes the WBCELoss with the given parameters.
        """
        super(WBCELoss, self).__init__(**kwargs)

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor, mask=False, pos=1, neg=0.1
    ) -> torch.Tensor:
        """
        Computes the weighted BCE loss between the logits and targets.
        """
        logits, targets = self.initialize_tensors(logits, targets, mask)

        # Compute the percentage of positive samples
        positive_samples = torch.sum(targets)
        total_samples = targets.numel()
        positive_ratio = positive_samples / total_samples

        # Calculate class weights
        # TODO constant scale for pos and neg.
        # weight_positive = 1 / positive_ratio
        # weight_negative = 1 / (1 - positive_ratio)
        weight_positive = pos
        weight_negative = neg

        # Compute the binary cross entropy (BCE) loss
        bce_loss = -(
            (weight_positive * targets * torch.log(logits + 1e-8))
            + (weight_negative * (1 - targets) * torch.log(1 - logits + 1e-8))
        )

        # Average the losses across all samples
        if self.reduction == "sum":
            return torch.sum(bce_loss)
        elif self.reduction == "mean":
            return torch.mean(bce_loss)
        else:
            return bce_loss


class BCEMSELoss(AbstractLoss):
    """
    Implements the Binary Cross-Entropy over MSE loss function with an option
    to ignore the diagonal elements.

    The BCELoss class can be used for training where pixel-level accuracy is important.
    The MSE loos is used over continues Z slices to ensure smooth segmentation accuracy.
    """

    def __init__(self, mse_weight=0.1, **kwargs):
        """
        Initializes the BCELoss with the given parameters.
        """
        super(BCEMSELoss, self).__init__(**kwargs)

        self.mse_weight = mse_weight

        self.bce_loss = nn.BCELoss(reduction=self.reduction)
        self.mse_loss = nn.MSELoss(reduction=self.reduction)

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor, mask=True
    ) -> torch.Tensor:
        """
        Computes the BCE loss between the logits and targets.
        """
        logits, targets = self.initialize_tensors(logits, targets, mask)

        # Avg. MSE to past and future frame
        mse = (
            self.mse_loss(logits[:, :, :-1, ...], targets[:, :, 1:, ...])
            + self.mse_loss(logits[:, :, 1:, ...], targets[:, :, :-1])
        ) / 2

        # Regular BCE
        bce = self.bce_loss(logits, targets)

        return bce + (mse * self.mse_weight)



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

import struct
from collections import namedtuple
from os import listdir
from os.path import isfile, join
from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd
import tifffile.tifffile as tif

try:
    import nd2
except ImportError:
    pass

from numpy import ndarray
from plyfile import PlyData
from sklearn.neighbors import KDTree

from tardis_em.dist_pytorch.utils.utils import (
    RandomDownSampling,
    VoxelDownSampling,
)
from tardis_em.utils.errors import TardisError
from tardis_em.utils.normalization import RescaleNormalize
from tardis_em.utils import SCANNET_COLOR_MAP_20

import importlib.util as lib_utils


class ImportDataFromAmira:
    """
    LOADER FOR AMIRA SPATIAL GRAPH FILES

    This class read any .am file and if the spatial graph is recognized it is converted
    into a numpy array as (N, 4) with class ids and coordinates for XYZ.
    Also, due to Amira's design, file properties are encoded only in the image file
    therefore in order to properly ready spatial graph, class optionally requires
    amira binary or ASCII image file which contains transformation properties and
    pixel size. If the image file is not included, the spatial graph is returned without
    corrections.

    Args:
        src_am (str): Amira spatial graph directory.
        src_img (str, optional): Amira binary or ASCII image file directory.
    """

    def __init__(self, src_am: str, src_img: Optional[str] = None):
        self.src_img = src_img
        self.src_am = src_am

        # Read image and its property if existing
        if self.src_img is not None:
            if not self.src_img[-3:] == ".am":
                TardisError(
                    "130",
                    "tardis_em/utils/load_data.py",
                    f"{self.src_img} Not a .am file...",
                )

            if src_img.split("/")[-1:][:-3] != src_am.split("/")[-1:][:-20]:
                TardisError(
                    "131",
                    "tardis_em/utils/load_data.py",
                    f"Image file {src_img} has wrong extension for {src_am}!",
                )

            try:
                # Image file [Z x Y x X]
                (
                    self.image,
                    self.pixel_size,
                    self.physical_size,
                    self.transformation,
                ) = load_am(src_img)
            except RuntimeWarning:
                TardisError(
                    "130",
                    "tardis_em/utils/load_data.py",
                    "Directory or input .am image file is not correct..."
                    f"for given dir: {src_img}",
                )
        else:
            self.pixel_size = 1

        # Read spatial graph
        am = ""
        frame = 500
        while "# Data section follows" not in am:
            am = open(src_am, "r", encoding="iso-8859-1").read(frame)
            frame += 100

            if frame == 10000:
                break

        binary = False
        spatial_graph = ""
        if not any(
            [
                True
                for i in ["AmiraMesh 3D ASCII", "# ASCII Spatial Graph"]
                if i not in am
            ]
        ):
            if "AmiraMesh BINARY-LITTLE-ENDIAN 3.0" not in am:
                spatial_graph = None
            else:
                binary = True
        if spatial_graph is not None:
            self.spatial_graph = (
                open(src_am, "r", encoding="iso-8859-1").read().split("\n")
            )
            self.spatial_graph = [x for x in self.spatial_graph if x != ""]

        if binary:
            self.spatial_graph = None
            # self.spatial_graph = self.am_decode(self.spatial_graph)

    def __am_decode(self, am: str) -> str:
        pass

    def __get_segments(self) -> Union[np.ndarray, None]:
        """
        Helper class function to read segment data from amira file.

        Returns:
            np.ndarray: Array (N, 1) indicating a number of points per segment.
        """
        if self.spatial_graph is None:
            return None

        # Find line starting with EDGE { int NumEdgePoints }
        segments = str(
            [
                word
                for word in self.spatial_graph
                if word.startswith("EDGE { int NumEdgePoints }")
            ]
        )

        segment_start = "".join((ch if ch in "0123456789" else " ") for ch in segments)
        segment_start = [int(i) for i in segment_start.split()]

        # Find in the line directory that starts with @..
        try:
            segment_start = (
                int(self.spatial_graph.index("@" + str(segment_start[0]))) + 1
            )
        except ValueError:
            segment_start = (
                int(self.spatial_graph.index("@" + str(segment_start[0]) + " ")) + 1
            )

        # Find line define EDGE ... <- number indicate number of segments
        segments = str(
            [word for word in self.spatial_graph if word.startswith("define EDGE")]
        )

        segment_finish = "".join((ch if ch in "0123456789" else " ") for ch in segments)
        segment_finish = [int(i) for i in segment_finish.split()]
        segment_no = int(segment_finish[0])
        segment_finish = segment_start + int(segment_finish[0])

        # Select all lines between @.. (+1) and number of segments
        segments = self.spatial_graph[segment_start:segment_finish]
        segments = [i.split(" ")[0] for i in segments]

        # return an array of number of points belonged to each segment
        segment_list = np.zeros((segment_no, 1), dtype="int")
        segment_list[0:segment_no, 0] = [int(i) for i in segments]

        return segment_list

    def __find_points(self) -> Union[np.ndarray, None]:
        """
        Helper class function to search for points in Amira file.

        Returns:
            np.ndarray: Set of all points.
        """
        if self.spatial_graph is None:
            return None

        # Find line starting with POINT { float[3] EdgePointCoordinates }
        points = str(
            [
                word
                for word in self.spatial_graph
                if word.startswith("POINT { float[3] EdgePointCoordinates }")
            ]
        )

        # Find in the line directory that starts with @..
        points_start = "".join((ch if ch in "0123456789" else " ") for ch in points)
        points_start = [int(i) for i in points_start.split()]
        # Find line that start with the directory @... and select last one
        try:
            points_start = int(self.spatial_graph.index("@" + str(points_start[1]))) + 1
        except ValueError:
            points_start = (
                int(self.spatial_graph.index("@" + str(points_start[1]) + " ")) + 1
            )

        # Find line define POINT ... <- number indicate number of points
        points = str(
            [word for word in self.spatial_graph if word.startswith("define POINT")]
        )

        points_finish = "".join((ch if ch in "0123456789" else " ") for ch in points)
        points_finish = [int(i) for i in points_finish.split()][0]
        points_no = points_finish
        points_finish = points_start + points_finish

        # Select all lines between @.. (-1) and number of points
        points = self.spatial_graph[points_start:points_finish]

        # return an array of all points coordinates in pixel
        point_list = np.zeros((points_no, 3), dtype="float")
        for j in range(3):
            coord = [i.split(" ")[j] for i in points]
            point_list[0:points_no, j] = [float(i) for i in coord]

        return point_list

    def get_points(self) -> Union[np.ndarray, None]:
        """
        General class function to retrieve point cloud.

        Returns:
            np.ndarray: Point cloud as [X, Y, Z] after transformation and
                pixel size correction.
        """
        if self.spatial_graph is None:
            return None

        if self.src_img is None:
            self.transformation = [0, 0, 0]
        points_coord = self.__find_points()

        points_coord[:, 0] = points_coord[:, 0] - self.transformation[0]
        points_coord[:, 1] = points_coord[:, 1] - self.transformation[1]
        points_coord[:, 2] = points_coord[:, 2] - self.transformation[2]

        try:
            coordinate = str(
                [
                    word
                    for word in self.spatial_graph
                    if word.startswith("        Coordinates")
                ]
            ).split(" ")[9][1:-3]

            if coordinate == "nm":
                return points_coord / (self.pixel_size / 10)
        except IndexError:
            pass

        return points_coord / self.pixel_size

    def get_segmented_points(self) -> Union[np.ndarray, None]:
        """
        General class function to retrieve segmented point cloud.

        Returns:
            np.ndarray:  Point cloud as [ID, X, Y, Z].
        """
        if self.spatial_graph is None:
            return None

        points = self.get_points()
        segments = self.__get_segments()

        segmentation = np.zeros((points.shape[0],))
        id_ = 0
        idx = 0
        for i in segments:
            segmentation[id_ : (id_ + int(i))] = idx

            idx += 1
            id_ += int(i)

        return np.stack((segmentation, points[:, 0], points[:, 1], points[:, 2])).T

    def get_labels(self) -> Union[dict, None]:
        """
        General class function to read all labels from amira file.

        Returns:
            dict: Dictionary with label IDs
        """
        if self.spatial_graph is None:
            return None

        # Find line starting with EDGE { int NumEdgePoints } associated with all labels
        labels = [
            word
            for word in self.spatial_graph
            if word.startswith("EDGE { int ")
            and not word.startswith("EDGE { int NumEdgePoints }")
        ]

        # Find line define EDGE ... <- number indicate number of segments
        segment_no = str(
            [word for word in self.spatial_graph if word.startswith("define EDGE")]
        )

        labels_dict = {}
        for i in labels:
            # Find line starting with EDGE { int label }
            label_start = "".join((ch if ch in "0123456789" else " ") for ch in i)
            label_start = [int(i) for i in label_start.split()][-1:]

            # Find in the line directory that starts with @..
            try:
                label_start = (
                    int(self.spatial_graph.index("@" + str(label_start[0]))) + 1
                )
            except ValueError:
                label_start = (
                    int(self.spatial_graph.index("@" + str(label_start[0]) + " ")) + 1
                )

            label_finish = "".join(
                (ch if ch in "0123456789" else " ") for ch in segment_no
            )
            label_finish = [int(i) for i in label_finish.split()]

            label_no = int(label_finish[0])
            label_finish = label_start + int(label_finish[0])

            # Select all lines between @.. (+1) and number of segments
            label = self.spatial_graph[label_start:label_finish]
            label = [i.split(" ")[0] for i in label]

            # return an array of number of points belonged to each segment
            label_list = np.zeros((label_no, 1), dtype="int")
            label_list[0:label_no, 0] = [int(i) for i in label]
            label_list = np.where(label_list != 0)[0]

            labels_dict.update({i[11:-5].replace(" ", "").replace("}", ""): label_list})

        return labels_dict

    def get_image(self):
        """
        General class function to return image file.

        Returns:
            np.ndarray, float: Image and if available pixel size data.
        """
        return self.image, self.pixel_size

    def get_pixel_size(self) -> float:
        """
        Catch pixel size value from image.

        Returns:
            float: Pixel size.
        """
        return self.pixel_size


def load_tiff(tiff: str) -> Tuple[np.ndarray, float]:
    """
    Function to load any tiff file.

    Args:
        tiff (str): Tiff file directory.

    Returns:
        np.ndarray, float: Image data and unified pixel size.
    """
    if not isfile(tiff):
        TardisError(
            "130",
            "tardis_em/utils/load_data.py",
            f"Indicated .tif  {tiff} file does not exist...",
        )

    return np.array(tif.imread(tiff)), 1.0


# int nx
# int ny
# int nz
fstr = "3i"
names = "nx ny nz"

# int mode
fstr += "i"
names += " mode"

# int nxstart
# int nystart
# int nzstart
fstr += "3i"
names += " nxstart nystart nzstart"

# int mx
# int my
# int mz
fstr += "3i"
names += " mx my mz"

# float xlen
# float ylen
# float zlen
fstr += "3f"
names += " xlen ylen zlen"

# float alpha
# float beta
# float gamma
fstr += "3f"
names += " alpha beta gamma"

# int mapc
# int mapr
# int maps
fstr += "3i"
names += " mapc mapr maps"

# float amin
# float amax
# float amean
fstr += "3f"
names += " amin amax amean"

# int ispg
# int next
# short creatid
fstr += "2ih"
names += " ispg next creatid"

# pad 30 (extra data)
# [98:128]
fstr += "30x"

# short nint
# short nreal
fstr += "2h"
names += " nint nreal"

# pad 20 (extra data)
# [132:152]
fstr += "20x"

# int imodStamp
# int imodFlags
fstr += "2i"
names += " imodStamp imodFlags"

# short idtype
# short lens
# short nd1
# short nd2
# short vd1
# short vd2
fstr += "6h"
names += " idtype lens nd1 nd2 vd1 vd2"

# float[6] tiltangles
fstr += "6f"
names += " tilt_ox tilt_oy tilt_oz tilt_cx tilt_cy tilt_cz"

# NEW-STYLE MRC image2000 HEADER - IMOD 2.6.20 and above
# float xorg
# float yorg
# float zorg
# char[4] cmap
# char[4] stamp
# float rms
fstr += "3f4s4sf"
names += " xorg yorg zorg cmap stamp rms"

# int nlabl
# char[10][80] labels
fstr += "i800s"
names += " nlabl labels"

header_struct = struct.Struct(fstr)
MRCHeader = namedtuple("MRCHeader", names)


def mrc_read_header(mrc: Union[str, bytes, None] = None):
    """
    Helper function to read MRC header.

    Args:
        mrc (str): MRC file directory.

    Returns:
        class: MRC header.
    """
    if isinstance(mrc, str):
        with open(mrc, "rb") as f:
            header = f.read(1024)
    else:
        header = mrc

    return MRCHeader._make(header_struct.unpack(header))


def mrc_write_header(*args) -> bytes:
    header = MRCHeader(*args)
    return header_struct.pack(*list(header))


def mrc_mode(mode: int, amin: int):
    """
    Helper function to decode MRC mode type.

    mode int: MRC mode from mrc header.
    amin int: MRC minimum pixel value.

    Returns:
        np.dtype: Mode as np.dtype.
    """
    dtype_ = {
        0: np.uint8,
        1: np.int16,  # Signed 16-bit integer
        2: np.float32,  # Signed 32-bit real
        3: "2h",  # Complex 16-bit integers
        4: np.complex64,  # Complex 32-bit reals
        6: np.uint16,  # Unassigned int16
        12: np.float16,  # Signed 16-bit half-precision real
        16: "3B",  # RGB values
    }

    if mode == 101:
        TardisError(
            "130",
            "tardis_em/utils/load_data.py",
            "4 bit .mrc file are not supported. Ask Dev if you need it!",
        )
    if mode == 1024:
        TardisError(
            "130",
            "tardis_em/utils/load_data.py",
            "Are your trying to load tiff file as mrc?",
        )

    if isinstance(mode, int):
        if mode == 0 and amin >= 0:
            return dtype_[mode]
        elif mode == 0 and amin < 0:
            return np.int8

        if mode in dtype_:
            return dtype_[mode]
        else:
            TardisError(
                "130",
                "tardis_em/utils/load_data.py",
                f"Unknown dtype mode: {str(mode)} and {str(amin)}",
            )
    else:
        if mode in [np.int8, np.uint8]:
            return 0
        for name in dtype_:
            if mode == dtype_[name]:
                return name


def load_am(am_file: str):
    """
    Function to load Amira binary image data.

    Args:
        am_file (str): Amira binary image .am file.

    Returns:
        np.ndarray, float, float, list: Image file as well images parameters.
    """
    if not isfile(am_file):
        TardisError(
            "130",
            "tardis_em/utils/load_data.py",
            f"Indicated .am {am_file} file does not exist...",
        )

    am = open(am_file, "r", encoding="iso-8859-1").read(8000)

    asci = False
    if "AmiraMesh 3D ASCII" in am:
        if "define Lattice" not in am:
            TardisError(
                "130",
                "tardis_em/utils/load_data.py",
                f".am {am_file} file is coordinate file not image!",
            )
        asci = True

    size = [word for word in am.split("\n") if word.startswith("define Lattice ")][0][
        15:
    ].split(" ")

    nx, ny, nz = int(size[0]), int(size[1]), int(size[2])

    # Fix for ET that were trimmed
    #  ET boundary box has wrong size
    bb = str(
        [word for word in am.split("\n") if word.startswith("    BoundingBox")]
    ).split(" ")

    if len(bb) == 0:
        physical_size = np.array((float(bb[6]), float(bb[8]), float(bb[10][:-3])))
        transformation = np.array((0.0, 0.0, 0.0))
    else:
        am = open(am_file, "r", encoding="iso-8859-1").read(20000)
        bb = str(
            [word for word in am.split("\n") if word.startswith("    BoundingBox")]
        ).split(" ")

        physical_size = np.array((float(bb[6]), float(bb[8]), float(bb[10][:-3])))

        transformation = np.array((float(bb[5]), float(bb[7]), float(bb[9])))

    try:
        coordinate = str(
            [word for word in am.split("\n") if word.startswith("        Coordinates")]
        ).split(" ")[9][1:-3]
    except IndexError:
        coordinate = None

    if coordinate == "m":  # Bring meter to angstrom
        pixel_size = ((physical_size[0] - transformation[0]) / (nx - 1)) * 10000000000
    elif coordinate == "nm":  # Bring nm to angstrom
        pixel_size = ((physical_size[0] - transformation[0]) / (nx - 1)) * 10
    else:
        pixel_size = (physical_size[0] - transformation[0]) / (nx - 1)

    physical_size = (physical_size[0] - transformation[0]) / (nx - 1)
    pixel_size = np.round(pixel_size, 3)

    if "Lattice { byte Data }" in am or "Lattice { float Data }" in am:
        if asci:
            img = (
                open("../../rand_sample/T216_grid3b.am", "r", encoding="iso-8859-1")
                .read()
                .split("\n")
            )
            img = [x for x in img if x != ""]
            img = np.asarray(img)
            return img
        else:
            if "Lattice { byte Data }" in am:
                dtype_ = np.uint8
                img = np.fromfile(am_file, dtype=dtype_)
            else:
                dtype_ = np.float32
                offset = str.find(am, "\n@1\n") + 4
                img = np.fromfile(am_file, dtype=np.float32, offset=offset)
    elif "Lattice { sbyte Data }" in am:
        dtype_ = np.int8
        img = np.fromfile(am_file, dtype=dtype_)
        img = img + 128
    else:
        TardisError(
            "130",
            "tardis_em/utils/load_data.py",
            f".am {am_file} not supported .am format!",
        )
        return None

    if dtype_ != np.float32:
        binary_start = str.find(am, "\n@1\n") + 4
        img = img[binary_start:-1]
    if nz == 1:
        if len(img) == ny * nx:
            img = img.reshape((ny, nx))
        else:
            df_img = np.zeros((ny * nx), dtype=dtype_)
            df_img[: len(img)] = img
            img = df_img.reshape((ny, nx))
    else:
        if len(img) == nz * ny * nx:
            img = img.reshape((nz, ny, nx))
            # if dtype_ == np.float32:
            # img = np.flip(img, 0)
        else:
            df_img = np.zeros((nz * ny * nx), dtype=dtype_)
            df_img[: len(img)] = img
            img = df_img.reshape((nz, ny, nx))

    return img, pixel_size, physical_size, transformation


def load_mrc_file(mrc: str) -> Union[Tuple[np.ndarray, float], Tuple[None, float]]:
    """
    Function to load MRC 2014 file format.

    Args:
        mrc (str): MRC file directory.

    Returns:
        np.ndarray, float: Image data and pixel size.
    """
    if not isfile(mrc):
        TardisError(
            "130",
            "tardis_em/utils/load_data.py",
            f"Indicated .mrc {mrc} file does not exist...",
        )

    header = mrc_read_header(mrc)
    extended_header = header.next

    pixel_size = round(header.xlen / header.nx, 3)
    dtype = mrc_mode(header.mode, header.amin)
    nz, ny, nx = header.nz, header.ny, header.nx
    bit_len = nz * ny * nx

    # Check for corrupted files
    try:
        if nz == 1:
            image = np.fromfile(mrc, dtype=dtype)[-bit_len:].reshape((ny, nx))
        else:
            image = np.fromfile(mrc, dtype=dtype)[-bit_len:].reshape((nz, ny, nx))
    except ValueError:  # File is corrupted try to load as much as possible
        if nz > 1:
            if mrc.endswith(".rec"):
                header_len = 512
            else:
                header_len = 1024 + extended_header
            image = np.fromfile(mrc, dtype=dtype)[header_len:]

            while bit_len >= len(image):
                nz = nz - 1
                bit_len = nz * ny * nx

            image = image[:bit_len]
            image = image.reshape((nz, ny, nx))
        else:
            image = None

    if image is None:
        return None, 1.0

    # if image.min() < 0 and image.dtype == np.int8:
    #     image = image + 128
    #     image = image.astype(np.uint8)
    #
    # if image.min() < 0 and image.dtype == np.int16:
    #     image = image + 32768
    #     image = image.astype(np.uint16)

    # Detect wrongly saved cryo-EM mrc files
    if nz > int(ny * 2):
        image = image.transpose((1, 0, 2))  # YZX to ZYX
    elif nz > int(nx * 2):
        image = image.transpose((2, 1, 0))  # XYZ to ZYX

    return image, pixel_size


def load_nd2_file(
    nd2_dir: str, channels=True
) -> Union[Tuple[np.ndarray, float], Tuple[None, float]]:
    try:
        nd2.version("nd2")
    except NameError:
        TardisError(
            id_="2",
            py="tardis_em/utils/load_data.py",
            desc="You are trying to load ND2 file. Please install nd2 library or "
            'install tardis with nd2 by: pip install "tardis_em[nd2]"',
        )
        return None, 1.0

    img = nd2.imread(nd2_dir)

    if channels:
        # Search for the smallest dim, which will indicate channels
        dim0 = img[0, 0, ...]
        dim0 = int(np.round(np.mean(dim0) / np.std(dim0), 0))

        dim1 = img[0, 1, ...]
        dim1 = int(np.round(np.mean(dim1) / np.std(dim1), 0))

        dim2 = img[1, 0, ...]
        dim2 = int(np.round(np.mean(dim2) / np.std(dim2), 0))

        dim3 = img[1, 1, ...]
        dim3 = int(np.round(np.mean(dim3) / np.std(dim3), 0))

        if dim0 >= dim2 and dim0 > dim3 and dim1 > dim2 and dim1 >= dim3:
            img = img[0 if dim0 > dim2 else 1, ...]
        else:
            img = img[:, 0 if dim0 > dim1 else 1, ...]

    return img, 1.0


def load_ply_scannet(
    ply: str, downscaling=0, color: Optional[str] = None
) -> Union[Tuple[ndarray, ndarray], ndarray]:
    """
    Function to read .ply files.
    Args:
        ply (str): File directory.
        downscaling (float): Downscaling point cloud by fixing voxel size defaults to 0.1.
        color (str, optional): Optional color feature defaults to None.
    Returns:
        np.ndarray: Label point cloud coordinates and optionally RGB value for
            each point.
    """
    # Load .ply scannet file
    ply = PlyData.read(ply)["vertex"]

    pcd = np.stack(
        (ply["x"], ply["y"], ply["z"], ply["red"], ply["green"], ply["blue"]), axis=-1
    ).astype(np.float32)

    coord = pcd[:, :3]
    label = pcd[:, 3:]

    # Retrieve ScanNet v2 labels after down scaling
    cls_id = np.zeros((len(label), 1))
    get_key_from_value = lambda value: next(
        (key for key, val in SCANNET_COLOR_MAP_20.items() if val == value), None
    )

    for id_, i in enumerate(label):
        cls_id[id_, 0] = get_key_from_value(tuple(i))
    coord = np.hstack((cls_id, coord))

    # Retrieve Node RGB features
    if color is not None:
        ply = PlyData.read(color)["vertex"]

        rgb = np.stack((ply["red"], ply["green"], ply["blue"]), axis=-1).astype(
            np.float32
        )

        if downscaling > 0:
            down_scale = VoxelDownSampling(voxel=downscaling, labels=True)
            coord, rgb = down_scale(coord=coord, rgb=rgb)

        if coord[:, 1:].shape != rgb.shape:
            TardisError(
                "131",
                "tardis_em/utils/load_data.py",
                "RGB shape must be the same as coord! "
                f"But {coord[:, 1:].shape} != {rgb.shape}",
            )
        return coord, rgb
    else:
        # Down scaling point cloud with labels
        if downscaling > 0:
            down_scale = VoxelDownSampling(voxel=downscaling, labels=True)
            coord = down_scale(coord)
        return coord


def load_ply_partnet(ply, downscaling=0) -> np.ndarray:
    """
    Function to read .ply files.
    Args:
        ply (str): File directory.
        downscaling (float): Downscaling point cloud by fixing voxel size.
    Returns:
        np.ndarray: Labeled point cloud coordinates.
    """
    # Load .ply scannet file
    ply = PlyData.read(ply)["vertex"]

    pcd = np.stack(
        (ply["x"], ply["y"], ply["z"], ply["red"], ply["green"], ply["blue"]), axis=-1
    ).astype(np.float32)

    coord_org = pcd[:, :3]
    label_org = pcd[:, 3:]
    label_uniq = np.unique(label_org, axis=0)

    # Downscaling point cloud with labels
    down_scale = VoxelDownSampling(voxel=downscaling, labels=False)
    if downscaling > 0:
        coord = down_scale(coord_org)
    else:
        coord = coord_org

    label_id = []
    tree = KDTree(coord_org, leaf_size=coord_org.shape[0])
    for i in coord:
        _, match_coord = tree.query(i.reshape(1, -1), k=1)
        match_coord = match_coord[0][0]

        label_id.append(np.where(np.all(label_org[match_coord] == label_uniq, 1))[0][0])

    return np.hstack((np.asarray(label_id)[:, None], coord))


def load_txt_s3dis(
    txt: str, rgb=False, downscaling=0
) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
    """
    Function to read .txt Stanford 3D instance scene file.

    Args:
        txt (str): File directory.
        rgb (bool): If True return RGB value.
        downscaling (float): Downscaling point cloud by fixing voxel size.

    Returns:
        np.ndarray: Labeled point cloud coordinates.
    """
    coord = pd.read_csv(txt, sep=" ", on_bad_lines="skip").to_numpy()

    rgb_values = coord[:, 3:]
    coord = coord[:, :3]

    # Downscaling point cloud with labels
    down_scale = VoxelDownSampling(voxel=downscaling, labels=False)
    if downscaling > 0:
        coord = down_scale(coord=coord)

        if rgb:
            coord, rgb_values = down_scale(coord=coord, rgb=rgb_values)

    if rgb:
        return coord, rgb_values
    return coord


def load_s3dis_scene(
    dir_: str, downscaling=0, random_ds=None, rgb=False
) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
    """
    Function to read .txt Stanford 3D instance scene files.

    Args:
        dir_ (str): Folder directory with all instances.
        downscaling (float): Downscaling point cloud by fixing voxel size.
        random_ds (None, float): If not None, indicate ration of point to keep.
        rgb (bool): If True, load rgb value.

    Returns:
        np.ndarray: Labeled point cloud coordinates.
    """
    dir_list = [x for x in listdir(dir_) if x not in [".DS_Store", "Icon"]]

    # Build S3DIS scene with IDs [ID, X, Y, Z] [R, G, B]
    coord_scene = []
    rgb_scene = []
    id_ = 0
    for i in dir_list:
        if rgb:
            coord_inst, rgb_v = load_txt_s3dis(join(dir_, i), rgb=rgb)
            rgb_scene.append(rgb_v)
        else:
            coord_inst = load_txt_s3dis(join(dir_, i))

        coord_scene.append(
            np.hstack((np.repeat(id_, len(coord_inst)).reshape(-1, 1), coord_inst))
        )

        id_ += 1
    coord = np.concatenate(coord_scene)
    if rgb:
        rgb_v = np.concatenate(rgb_scene) / 255
    else:
        rgb_v = None

    # Down scale scene
    if downscaling > 0 or random_ds is not None:
        if random_ds is not None:
            down_scale = RandomDownSampling(threshold=random_ds, labels=True)
        else:
            down_scale = VoxelDownSampling(voxel=downscaling, labels=True)

        if rgb:
            return down_scale(coord=coord, rgb=rgb_v)
        else:
            return down_scale(coord=coord)


def load_image(
    image: str,
    normalize=False,
    px_=True,
) -> Union[np.ndarray, Tuple[np.ndarray, float]]:
    """
    Quick wrapper for loading image data based on the detected file format.

    Args:
        image (str): Image file directory.
        normalize (bool): Rescale histogram to 1% - 99% percentile.
        px_ (bool): Return px if True

    Returns:
        np.ndarray, float: Image array and associated pixel size.
    """
    px = 0.0

    if image.endswith((".tif", ".tiff")):
        image, _ = load_tiff(image)
        px = 1.0
    elif image.endswith((".mrc", ".rec", ".map")):
        image, px = load_mrc_file(image)
    elif image.endswith(".am"):
        image, px, _, _ = load_am(image)
    elif image.endswith(".nd2"):
        library_names = ["nd2"]
        library_installed = False
        for lib in library_names:
            if lib_utils.find_spec(lib) is not None:
                library_installed = True

        assert library_installed, (
            f"ND2 library is not installed." f'Run "pip install np2" to install it!'
        )

        image, px = load_nd2_file(image, channels=True)

    if normalize:
        norm = RescaleNormalize(clip_range=(1, 99))
        image = norm(image)

    if isinstance(image, str):
        TardisError(
            "130",
            "tardis_em/utils/load_data.py",
            f"Indicated .mrc {image} file does not exist...",
        )

    if px_:
        return image, px
    return image



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from typing import Optional, Tuple, Union

import torch
import torch.nn as nn

from tardis_em.dist_pytorch.model.embedding import EdgeEmbedding, NodeEmbedding
from tardis_em.dist_pytorch.model.layers import DistStack
from tardis_em.utils.errors import TardisError


class BasicDIST(nn.Module):
    """
    General DIST FOR DIMENSIONLESS INSTANCE SEGMENTATION TRANSFORMER

    Args:
        n_out (int): Number of channels in the output layer.
        node_input (int): Length of the flattened image file.
        node_dim (int, None): In features of image for linear transformation.
        edge_dim (int): In feature of coord for linear transformation.
        num_layers (int): Number of DIST layers to initialize.
        num_heads (int): Number of heads for MHA.
        num_cls (int, None): Number of predicted classes.
        coord_embed_sigma (float): Sigma value used to embed coordinate distance features.
        dropout_rate (float): Dropout factor used in MHA dropout layer.
        structure (str): DIST network structure. (full, triang, dualtriang, quad, attn)
        predict (bool): If True sigmoid output.
    """

    def __init__(
        self,
        n_out=1,
        node_input=0,
        node_dim=None,
        edge_dim=128,
        num_layers=6,
        num_heads=8,
        num_cls=None,
        rgb_embed_sigma=1.0,
        coord_embed_sigma=1.0,
        dropout_rate=0,
        structure="full",
        predict=False,
        edge_angles=False,
    ):
        super(BasicDIST, self).__init__()

        self.n_out = n_out
        self.node_input = node_input
        self.node_dim = node_dim
        self.edge_dim = edge_dim
        self.num_layers = num_layers
        self.num_heads = num_heads
        self.num_cls = num_cls
        self.node_sigma = rgb_embed_sigma
        self.edge_sigma = coord_embed_sigma
        self.dropout_rate = dropout_rate
        self.structure = structure
        self.predict = predict

        if self.node_dim is not None:
            if self.node_input > 0:
                self.node_embed = NodeEmbedding(
                    n_in=self.node_input, n_out=self.node_dim, sigma=self.node_sigma
                )

        self.coord_embed = EdgeEmbedding(n_out=self.edge_dim, sigma=self.edge_sigma)

        self.layers = DistStack(
            node_dim=self.node_dim,
            pairs_dim=self.edge_dim,
            dropout=self.dropout_rate,
            num_layers=self.num_layers,
            num_heads=self.num_heads,
            structure=self.structure,
        )

        self.decoder = nn.Linear(in_features=self.edge_dim, out_features=self.n_out)

        if self.predict:
            self.logits_sigmoid = nn.Sigmoid()

    def embed_input(
        self, coords: torch.Tensor, node_features: Optional[torch.Tensor] = None
    ):
        """
        Embedding features

        Args:
            coords (torch.Tensor): Coordinate features.
            node_features (torch.Tensor, None): Optional Node features.

        Returns:
            torch.tensor: Embedded features for prediction.
        """
        if node_features is not None:
            """Batch x Length x Embedded_Dim"""
            x = self.node_embed(input_node=node_features)
        else:
            x = None

        """ Batch x Length x Length x Channels """
        z = self.coord_embed(input_coord=coords)
        return x, z

    def forward(
        self, coords: torch.Tensor, node_features=None
    ) -> Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:
        """
        Forward DIST model.

        Args:
            coords (torch.Tensor): Coordinates input of a shape
                [Batch x Length x Channels].
            node_features (torch.Tensor, None): Image patch input of a shape
                [Batch x Length x Channels].
        """
        node, edge = self.embed_input(coords=coords, node_features=node_features)

        if node is not None:
            """Length x Batch x Embedded_Dim"""
            node = node.transpose(0, 1)

        """ Encode throughout the transformer layers """
        _, edge = self.layers(
            node_features=node, edge_features=edge, src_key_padding_mask=None
        )

        """ Predict the graph edges """
        logits = self.decoder(edge + edge.transpose(1, 2))  # symmetries z
        logits = logits.permute(0, 3, 1, 2)

        # if self.num_cls is not None:
        #     # Batch x Length x Channels
        #     diag = np.arange(logits.shape[2])
        #     logits_cls = self.decoder_cls(edge)[:, diag, diag, :]

        if self.predict:
            # if self.num_cls is not None:
            #     # Batch x Length x Channel
            #     logits_cls = self.logits_cls_softmax(logits_cls)
            #     # Batch x Length
            #     logits_cls = torch.argmax(logits_cls, 2)
            # else:
            # Batch x Channels x Length x Length
            logits = self.logits_sigmoid(logits)

        # if self.num_cls is not None:
        #     return logits, logits_cls
        # else:
        return logits


class DIST(BasicDIST):
    """
    MAIN DIST FOR DIMENSIONLESS INSTANCE SEGMENTATION TRANSFORMER

    This transformer taking into the account the positional encoding of each
    coordinate point and attend them with patch image to which this coordinate
    is corresponding. This attention is aiming to training the transformer in
    outputting a graph from which point cloud can be segmented.


    Returns:
        torch.Tensor: DIST prediction after sigmoid (prediction) or last
            linear layer (training).
    """

    def __init__(self, **kwargs):
        super(DIST, self).__init__(**kwargs)


class CDIST(BasicDIST):
    """
    MAIN DIST FOR CLASSIFYING DIMENSIONLESS INSTANCE SEGMENTATION TRANSFORMER

    This transformer taking into the account the positional encoding of each
    coordinate point and attend them with patch image to which this coordinate
    is corresponding. This attention is aiming to training the transformer in
    outputting a graph from which point cloud can be segmented.

    Returns:
        torch.Tensor: DIST prediction as well as DIST class prediction, after
            sigmoid (prediction) or last linear layer (training).
    """

    def __init__(self, **kwargs):
        super(CDIST, self).__init__(**kwargs)
        if self.num_cls is None:
            TardisError(
                "build_cdist_network",
                "tardis_em/dist",
                "Undefined num_cls parameter!",
            )

        self.decoder_cls = nn.Linear(
            in_features=self.edge_dim, out_features=self.num_cls
        )

        if self.predict:
            self.logits_sigmoid = nn.Sigmoid()
        self.logits_cls_softmax = nn.Softmax(dim=2)


def build_dist_network(network_type: str, structure: dict, prediction: bool):
    """
    Wrapper for building DIST model

    Wrapper take DIST parameter and predefined network type (e.g. DIST, C_DIST),
    and build DIST model.

    Args:
        network_type (str): Network type name.
        structure (dict):  Dictionary with all network setting.
        prediction (bool): If True, build network in prediction path.

    Returns:
        DIST: DIST network structure.
    """
    if network_type not in ["instance", "semantic"]:
        TardisError(
            "build_dist_network",
            "tardis_em/dist",
            f"Wrong DIST network name {network_type}",
        )

    if network_type == "instance":
        return DIST(
            n_out=structure["n_out"],
            node_input=structure["node_input"],
            node_dim=structure["node_dim"],
            edge_dim=structure["edge_dim"],
            num_layers=structure["num_layers"],
            num_heads=structure["num_heads"],
            rgb_embed_sigma=structure["rgb_embed_sigma"],
            coord_embed_sigma=structure["coord_embed_sigma"],
            dropout_rate=structure["dropout_rate"],
            structure=structure["structure"],
            predict=prediction,
        )
    elif network_type == "semantic":
        return CDIST(
            n_out=structure["n_out"],
            node_input=structure["node_input"],
            node_dim=structure["node_dim"],
            edge_dim=structure["edge_dim"],
            num_layers=structure["num_layers"],
            num_heads=structure["num_heads"],
            num_cls=structure["num_cls"],
            coord_embed_sigma=structure["coord_embed_sigma"],
            dropout_rate=structure["dropout_rate"],
            structure=structure["structure"],
            predict=prediction,
        )
    else:
        return None



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
from typing import Union

import numpy as np
from scipy.interpolate import splev, splprep


def curvature(coord: np.ndarray, tortuosity_=False) -> Union[float, tuple]:
    """
    Calculate spline curvature.

    Args:
        coord (np.ndarray): Coordinates for each unsorted point idx.
        tortuosity_ (bool): Optional, if True calculates the tortuosity.

    Returns:
        float: Spline tortuosity measured with tortuosity.
    """

    tck, u = splprep(coord.T, s=0)

    # Calculate the first, second, and third derivatives
    der1 = splev(u, tck, der=1)  # First derivative (velocity)
    der2 = splev(u, tck, der=2)  # Second derivative (acceleration)

    # Stack the derivatives for easier manipulation
    r1 = np.vstack(der1).T  # First derivative (velocity)
    r2 = np.vstack(der2).T  # Second derivative (acceleration)

    # Calculate curvature
    curvature = (
        np.linalg.norm(np.cross(r1, r2), axis=1) / np.linalg.norm(r1, axis=1) ** 3
    )

    if tortuosity_:
        # Calculate the curve length (arc length) using the fine points
        arc_length = np.sum(np.linalg.norm(np.diff(coord, axis=0), axis=1))

        # Calculate the straight-line distance between the start and end points
        start_point = coord[0]
        end_point = coord[-1]
        straight_line_distance = np.linalg.norm(end_point - start_point)

        # Calculate tortuosity
        tortuosity = arc_length / straight_line_distance

        return curvature, tortuosity
    return curvature


def curvature_list(
    coord: np.ndarray, tortuosity_=False, mean_=False
) -> Union[list, tuple]:
    """
    Calculate the curvature of all splines and return it as a list.

    Args:
        coord (np.ndarray): Coordinates for each unsorted point idx.
        tortuosity_ (bool): Optional, if True calculates the tortuosity.
        mean_ (bool): If true, return an average curvature.

    Returns:
        list: Spline curvature list.
    """

    spline_curvature_list, spline_tortuosity_list = [], []

    for i in np.unique(coord[:, 0]):
        points = coord[np.where(coord[:, 0] == i)[0], 1:]

        if len(points) > 5:
            if tortuosity_:
                c, t = curvature(points, tortuosity_=True)

                if mean_:
                    c = np.mean(c).item(0)
                spline_curvature_list.append(c)
                spline_tortuosity_list.append(t)
            else:
                spline_curvature_list.append(curvature(points))
        else:
            spline_curvature_list.append(0.0)
            spline_tortuosity_list.append(1.0)

    if tortuosity_:
        return spline_curvature_list, [float(x) for x in spline_tortuosity_list]
    else:
        return spline_curvature_list


def tortuosity(coord: np.ndarray) -> float:
    """
    Calculate spline tortuosity.

    Args:
        coord (np.ndarray): Coordinates for each unsorted point idx.

    Returns:
        float: Spline tortuosity measured with tortuosity.
    """
    if len(coord) <= 1:
        return 1.0

    length = total_length(coord) + 1e-16
    start_point = coord[0]
    end_point = coord[-1]
    straight_line_distance = np.linalg.norm(end_point - start_point)

    return length / straight_line_distance


def tortuosity_list(coord: np.ndarray) -> list:
    """
    Calculate the tortuosity of all splines and return it as a list.

    Args:
        coord (np.ndarray): Coordinates for each unsorted point idx.

    Returns:
        list: Spline tortuosity list.
    """

    spline_tortuosity_list = []

    for i in np.unique(coord[:, 0]):
        points = coord[np.where(coord[:, 0] == i)[0], 1:]
        spline_tortuosity_list.append(tortuosity(points))

    return [float(x) for x in spline_tortuosity_list]


def total_length(coord: np.ndarray) -> float:
    """
    Calculate the total length of the spline.

    Args:
        coord (np.ndarray): Coordinates for each unsorted point idx.

    Returns:
        float: Spline length.
    """
    length = np.sum(np.linalg.norm(np.diff(coord, axis=0), axis=1))

    return length


def length_list(coord: np.ndarray) -> list:
    """
    Calculate the total length of all splines and return it as a list.

    Args:
        coord (np.ndarray): Coordinates for each unsorted point idx.

    Returns:
        list: Spline length list.
    """
    spline_length_list = []

    for i in np.unique(coord[:, 0]):
        points = coord[np.where(coord[:, 0] == i)[0], 1:]
        spline_length_list.append(total_length(points))

    return [float(x) for x in spline_length_list]


def angle_between_vectors(v1, v2):
    """
    Calculate the angle in degrees between two vectors.

    This function uses the dot product and the magnitudes of the vectors
    to calculate the angle between them according to the formula:

        cos(theta) = (A . B) / (||A|| ||B||)

    Args:
        v1(np.ndarray): First input vector.
        v2(np.ndarray): Second input vector.

    Returns:
        float The angle in degrees between vector 'v1' and 'v2'.
    """
    # Calculate the dot product of vectors v1 and v2
    dot_product = np.dot(v1, v2)

    # Calculate the magnitude (norm) of vectors
    magnitude_v1 = np.linalg.norm(v1) + 1e-16
    magnitude_v2 = np.linalg.norm(v2) + 1e-16

    # Calculate angle
    cos_theta = dot_product / (magnitude_v1 * magnitude_v2)
    angle = np.arccos(np.clip(cos_theta, -1.0, 1.0))

    return np.degrees(angle)


def intensity(data: np.ndarray, image: np.ndarray, thickness=1.0):
    tck, u = splprep(data.T, s=0)

    data_fine = np.linspace(0, 1, int(total_length(data)))
    data_fine = np.array(splev(data_fine, tck)).T

    pixel_coords = np.rint(data_fine).astype(int)

    # Extract the pixel values along the spline
    spline_intensity = pixel_intensity(pixel_coords, image)

    # Determined avg. background level
    spline_up = np.copy(pixel_coords)
    spline_up[:, 1] = spline_up[:, 1] + 10

    spline_down = np.copy(pixel_coords)
    spline_down[:, 1] = spline_down[:, 1] - 10

    # Extract the pixel values along the spline
    intensity_up = pixel_intensity(spline_up, image)
    intensity_down = pixel_intensity(spline_down, image)

    if intensity_up is None and intensity_down is not None:
        spline_background = intensity_down
    elif intensity_down is None and intensity_up is not None:
        spline_background = intensity_up
    elif intensity_up is not None and intensity_down is not None:
        if np.mean(intensity_up) > np.mean(intensity_down):
            spline_background = intensity_down
        else:
            spline_background = intensity_up
    else:
        return np.zeros_like(spline_intensity)
    return np.array(spline_intensity) - np.array(spline_background)


def pixel_intensity(coord, image):
    extracted_pixels = []

    if image.ndim == 2:
        for x, y, _ in coord:
            if 0 <= y < image.shape[0] and 0 <= x < image.shape[1]:
                extracted_pixels.append(image[y, x])
            else:
                extracted_pixels.append(None)
    else:
        for x, y, z in coord:
            if (
                0 <= z < image.shape[0]
                and 0 <= y < image.shape[1]
                and 0 <= x < image.shape[2]
            ):
                extracted_pixels.append(image[y, x])
            else:
                extracted_pixels.append(None)

    if None in extracted_pixels:
        try:
            min_value = min(item for item in extracted_pixels if item is not None)
            extracted_pixels = [
                min_value if item is None else item for item in extracted_pixels
            ]
        except ValueError:
            extracted_pixels = None
    return extracted_pixels


def intensity_list(coord: np.ndarray, image: np.ndarray):
    """
    Calculate the total length of all splines and return it as a list.

    Args:
        coord (np.ndarray): Coordinates for each unsorted point idx.
        image (np.ndarray): Associated image.

    Returns:
        list: Spline length list.
    """
    spline_intensity_list = []

    for i in np.unique(coord[:, 0]):
        points = coord[np.where(coord[:, 0] == i)[0], 1:]
        if len(points) > 5:
            spline_intensity_list.append(intensity(points, image))
        else:
            spline_intensity_list.append(0.0)

    return spline_intensity_list
